from typing import Callable, Iterable, Literal, Optional, overload

import hephaistos.pyhephaistos
import numpy.typing
import os

class AccelerationStructure:
    """
    Acceleration Structure used by programs to trace rays against a scene.
    Consists of multiple instances of various Geometries.

    Parameters
    ---------
    instances: GeometryInstance[]
        list of instances the structure consists of
    """

    def __init__(
        self, instances: hephaistos.pyhephaistos.GeometryInstanceVector
    ) -> None:
        """
        Creates an acceleration structure for consumption in shaders from the
        given geometry instances.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the acceleration structure to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the acceleration structure to the program at the given binding
        """
        ...

class AtomicsProperties:
    """
    List of atomic functions a device supports or are enabled
    """

    @property
    def bufferFloat16AtomicAdd(self) -> bool: ...
    @property
    def bufferFloat16AtomicMinMax(self) -> bool: ...
    @property
    def bufferFloat16Atomics(self) -> bool: ...
    @property
    def bufferFloat32AtomicAdd(self) -> bool: ...
    @property
    def bufferFloat32AtomicMinMax(self) -> bool: ...
    @property
    def bufferFloat32Atomics(self) -> bool: ...
    @property
    def bufferFloat64AtomicAdd(self) -> bool: ...
    @property
    def bufferFloat64AtomicMinMax(self) -> bool: ...
    @property
    def bufferFloat64Atomics(self) -> bool: ...
    @property
    def bufferInt64Atomics(self) -> bool: ...
    @property
    def imageFloat32AtomicAdd(self) -> bool: ...
    @property
    def imageFloat32AtomicMinMax(self) -> bool: ...
    @property
    def imageFloat32Atomics(self) -> bool: ...
    @property
    def imageInt64Atomics(self) -> bool: ...
    @property
    def sharedFloat16AtomicAdd(self) -> bool: ...
    @property
    def sharedFloat16AtomicMinMax(self) -> bool: ...
    @property
    def sharedFloat16Atomics(self) -> bool: ...
    @property
    def sharedFloat32AtomicAdd(self) -> bool: ...
    @property
    def sharedFloat32AtomicMinMax(self) -> bool: ...
    @property
    def sharedFloat32Atomics(self) -> bool: ...
    @property
    def sharedFloat64AtomicAdd(self) -> bool: ...
    @property
    def sharedFloat64AtomicMinMax(self) -> bool: ...
    @property
    def sharedFloat64Atomics(self) -> bool: ...
    @property
    def sharedInt64Atomics(self) -> bool: ...

class BindingTraits:
    """
    Properties of binding found in programs
    """

    @property
    def binding(self) -> int:
        """
        Index of the binding
        """
        ...
    @property
    def count(self) -> int:
        """
        Number of elements in binding, i.e. array size
        """
        ...
    @property
    def imageTraits(self) -> Optional[hephaistos.pyhephaistos.ImageBindingTraits]:
        """
        Properties of the image if one is expected
        """
        ...
    @property
    def name(self) -> str:
        """
        Name of the binding. Might be empty.
        """
        ...
    @property
    def type(self) -> hephaistos.pyhephaistos.ParameterType:
        """
        Type of the binding
        """
        ...

class Buffer:
    """
    Base class for all buffers managing memory allocation on the host
    """

    ...

class ByteBuffer:
    """
    Buffer representing memory allocated on the host holding an array of type
    uint8 and given size which can be accessed as numpy array.

    Parameters
    ----------
    size: int
        Number of elements
    """

    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array using this buffer's memory.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this buffer.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class ByteTensor:
    """
    Tensor representing memory allocated on the device holding an array of type
    uint8 and given size. If mapped can be accessed from the host using its memory
    address. Mapping can optionally be requested but will be ignored if the device
    does not support it. Query its support after creation via isMapped.
    """

    def __init__(self, array: numpy.typing.NDArray, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        array: NDArray
            Numpy array containing the data used to fill the tensor.
            Its type must match the tensor's.
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, size: int, mapped: bool = False) -> None:
        """
        Creates a new tensor of given size.

        Parameters
        ----------
        size: int
            Number of elements
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, addr: int, n: int, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        addr: int
            Address of the data used to fill the tensor
        n: int
            Number of bytes to copy from addr
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @property
    def address(self) -> int:
        """
        The device address of this tensor.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    def flush(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the host available to the device.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to flush
        size: int|None, default=None
            Number of elements to flush starting at offset.
            If None, flushes all remaining elements.
        """
        ...
    def invalidate(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the device available to the host.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to invalidate
        size: int | None, default=None
            Number of elements to invalidate starting at offset.
            If None, invalidates all remaining elements.
        """
        ...
    @property
    def isMapped(self) -> bool:
        """
        True, if the underlying memory is writable by the CPU.
        """
        ...
    @property
    def isNonCoherent(self) -> bool:
        """
        Wether calls to flush() and invalidate() are necessary to make changes
        in mapped memory between devices and host available
        """
        ...
    @property
    def memory(self) -> int:
        """
        Mapped memory address of the tensor as seen from the CPU. Zero if not mapped.
        """
        ...
    def retrieve(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Retrieves n elements from the tensor at the given offset and stores it
        in dst.

        Parameters
        ----------
        addr: int
            Address of memory to copy to
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this tensor.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the tensor in bytes.
        """
        ...
    def update(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Updates the tensor at the given offset with n elements stored at addr.

        Parameters
        ----------
        addr: int
            Address of memory to copy from
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...

class CharBuffer:
    """
    Buffer representing memory allocated on the host holding an array of type
    int8 and given size which can be accessed as numpy array.

    Parameters
    ----------
    size: int
        Number of elements
    """

    def __init__(self, size: int) -> None: ...
    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array using this buffer's memory.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this buffer.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class CharTensor:
    """
    Tensor representing memory allocated on the device holding an array of type
    int8 and given size. If mapped can be accessed from the host using its memory
    address. Mapping can optionally be requested but will be ignored if the
    device does not support it. Query its support after creation via isMapped.
    """

    def __init__(self, array: numpy.typing.NDArray, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        array: NDArray
            Numpy array containing the data used to fill the tensor.
            Its type must match the tensor's.
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, size: int, mapped: bool = False) -> None:
        """
        Creates a new tensor of given size.

        Parameters
        ----------
        size: int
            Number of elements
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, addr: int, n: int, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        addr: int
            Address of the data used to fill the tensor
        n: int
            Number of bytes to copy from addr
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @property
    def address(self) -> int:
        """
        The device address of this tensor.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    def flush(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the host available to the device.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to flush
        size: int|None, default=None
            Number of elements to flush starting at offset.
            If None, flushes all remaining elements.
        """
        ...
    def invalidate(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the device available to the host.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to invalidate
        size: int | None, default=None
            Number of elements to invalidate starting at offset.
            If None, invalidates all remaining elements.
        """
        ...
    @property
    def isMapped(self) -> bool:
        """
        True, if the underlying memory is writable by the CPU.
        """
        ...
    @property
    def isNonCoherent(self) -> bool:
        """
        Wether calls to flush() and invalidate() are necessary to make changes
        in mapped memory between devices and host available
        """
        ...
    @property
    def memory(self) -> int:
        """
        Mapped memory address of the tensor as seen from the CPU. Zero if not mapped.
        """
        ...
    def retrieve(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Retrieves n elements from the tensor at the given offset and stores it
        in dst.

        Parameters
        ----------
        addr: int
            Address of memory to copy to
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this tensor.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the tensor in bytes.
        """
        ...
    def update(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Updates the tensor at the given offset with n elements stored at addr.

        Parameters
        ----------
        addr: int
            Address of memory to copy from
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...

class ClearTensorCommand:
    """
    Command for filling a tensor with constant data over a given range
    """

    def __init__(
        self,
        tensor: hephaistos.pyhephaistos.Tensor,
        offset: Optional[int] = None,
        size: Optional[int] = None,
        data: Optional[int] = None,
        unsafe: bool = False,
    ) -> None:
        """
        Creates a command for filling a tensor with constant data over a given
        range. Defaults to zeroing the complete tensor
        """
        ...

class Command:
    """
    Base class for commands running on the device. Execution happens
    asynchronous after being submitted.
    """

class Compiler:
    """
    Compiler for generating SPIR-V byte code used by Programs from shader code
    written in GLSL. Has additional methods to handle includes.
    """

    def __init__(self) -> None: ...
    def addIncludeDir(self, path: os.PathLike) -> None:
        """
        Adds a path to the list of include directories that take part in
        resolving includes.
        """
        ...
    def clearIncludeDir(self) -> None:
        """
        Clears the internal list of include directories
        """
        ...
    def compile(self, code: str, headers: hephaistos.pyhephaistos.HeaderMap) -> bytes:
        """
        Compiles the given GLSL code using the provided header files and returns
        the SPIR-V code as bytes
        """
        ...
    @overload
    def compile(self, code: str) -> bytes:
        """
        Compiles the given GLSL code and returns the SPIR-V code as bytes
        """
        ...
    def popIncludeDir(self) -> None:
        """
        Removes the last added include dir from the internal list
        """
        ...

class DebugMessage:
    """
    Structure describing a debug message including text and some meta information
    """

    @property
    def idName(self) -> str:
        """
        Name of the message type
        """
        ...
    @property
    def idNumber(self) -> int:
        """
        Id of the message type
        """
        ...
    @property
    def message(self) -> str:
        """
        The actual message text
        """
        ...

class DebugMessageSeverityFlagBits:
    """
    Flags indicating debug message severity
    """

    INFO: DebugMessageSeverityFlagBits

    ERROR: DebugMessageSeverityFlagBits

    VERBOSE: DebugMessageSeverityFlagBits

    WARNING: DebugMessageSeverityFlagBits

class Device:
    """
    Handle for a physical device implementing the Vulkan API. Contains basic
    properties of the device.
    """

    @property
    def isDiscrete(self) -> bool:
        """
        True, if the device is a discrete GPU. Can be useful to distinguish from
        integrated ones.
        """
        ...
    @property
    def name(self) -> str:
        """
        Name of the device
        """
        ...

class DispatchCommand:
    """
    Command for executing a program using the given group size
    """

    @property
    def groupCountX(self) -> int:
        """
        Amount of groups in X dimension
        """
        ...
    @groupCountX.setter
    def groupCountX(self, arg: int, /) -> None:
        """
        Amount of groups in X dimension
        """
        ...
    @property
    def groupCountY(self) -> int:
        """
        Amount of groups in Y dimension
        """
        ...
    @groupCountY.setter
    def groupCountY(self, arg: int, /) -> None:
        """
        Amount of groups in Y dimension
        """
        ...
    @property
    def groupCountZ(self) -> int:
        """
        Amount of groups in Z dimension
        """
        ...
    @groupCountZ.setter
    def groupCountZ(self, arg: int, /) -> None:
        """
        Amount of groups in Z dimension
        """
        ...

class DispatchIndirectCommand:
    """
    Command for executing a program using the group size read from the provided tensor at given offset
    """

    @property
    def offset(self) -> int:
        """
        Offset into the Tensor in bytes on where to start reading
        """
        ...
    @offset.setter
    def offset(self, arg: int, /) -> None:
        """
        Offset into the Tensor in bytes on where to start reading
        """
        ...
    @property
    def tensor(self) -> Tensor:
        """
        Tensor from which to read the group size
        """
        ...
    @tensor.setter
    def tensor(self, arg: Tensor, /) -> None:
        """
        Tensor from which to read the group size
        """
        ...

class DoubleBuffer:
    """
    Buffer representing memory allocated on the host holding an array of type
    double and given size which can be accessed as numpy array.

    Parameters
    ----------
    size: int
        Number of elements
    """

    def __init__(self, size: int) -> None: ...
    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array using this buffer's memory.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this buffer.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class DoubleTensor:
    """
    Tensor representing memory allocated on the device holding an array of type
    double and given size. If mapped can be accessed from the host using its
    memory address. Mapping can optionally be requested but will be ignored if
    the device does not support it. Query its support after creation via isMapped.
    """

    def __init__(self, array: numpy.typing.NDArray, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        array: NDArray
            Numpy array containing the data used to fill the tensor.
            Its type must match the tensor's.
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, size: int, mapped: bool = False) -> None:
        """
        Creates a new tensor of given size.

        Parameters
        ----------
        size: int
            Number of elements
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, addr: int, n: int, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        addr: int
            Address of the data used to fill the tensor
        n: int
            Number of bytes to copy from addr
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @property
    def address(self) -> int:
        """
        The device address of this tensor.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    def flush(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the host available to the device.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to flush
        size: int|None, default=None
            Number of elements to flush starting at offset.
            If None, flushes all remaining elements.
        """
        ...
    def invalidate(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the device available to the host.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to invalidate
        size: int | None, default=None
            Number of elements to invalidate starting at offset.
            If None, invalidates all remaining elements.
        """
        ...
    @property
    def isMapped(self) -> bool:
        """
        True, if the underlying memory is writable by the CPU.
        """
        ...
    @property
    def isNonCoherent(self) -> bool:
        """
        Wether calls to flush() and invalidate() are necessary to make changes
        in mapped memory between devices and host available
        """
        ...
    @property
    def memory(self) -> int:
        """
        Mapped memory address of the tensor as seen from the CPU. Zero if not mapped.
        """
        ...
    def retrieve(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Retrieves n elements from the tensor at the given offset and stores it
        in dst.

        Parameters
        ----------
        addr: int
            Address of memory to copy to
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this tensor.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the tensor in bytes.
        """
        ...
    def update(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Updates the tensor at the given offset with n elements stored at addr.

        Parameters
        ----------
        addr: int
            Address of memory to copy from
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...

class FloatBuffer:
    """
    Buffer representing memory allocated on the host holding an array of type
    float and given size which can be accessed as numpy array.

    Parameters
    ----------
    size: int
        Number of elements
    """

    def __init__(self, size: int) -> None: ...
    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array using this buffer's memory.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this buffer.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class FloatTensor:
    """
    Tensor representing memory allocated on the device holding an array of type
    float and given size. If mapped can be accessed from the host using its
    memory address. Mapping can optionally be requested but will be ignored if
    the device does not support it. Query its support after creation via isMapped.
    """

    def __init__(self, array: numpy.typing.NDArray, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        array: NDArray
            Numpy array containing the data used to fill the tensor.
            Its type must match the tensor's.
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, size: int, mapped: bool = False) -> None:
        """
        Creates a new tensor of given size.

        Parameters
        ----------
        size: int
            Number of elements
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, addr: int, n: int, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        addr: int
            Address of the data used to fill the tensor
        n: int
            Number of bytes to copy from addr
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @property
    def address(self) -> int:
        """
        The device address of this tensor.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    def flush(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the host available to the device.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to flush
        size: int|None, default=None
            Number of elements to flush starting at offset.
            If None, flushes all remaining elements.
        """
        ...
    def invalidate(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the device available to the host.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to invalidate
        size: int | None, default=None
            Number of elements to invalidate starting at offset.
            If None, invalidates all remaining elements.
        """
        ...
    @property
    def isMapped(self) -> bool:
        """
        True, if the underlying memory is writable by the CPU.
        """
        ...
    @property
    def isNonCoherent(self) -> bool:
        """
        Wether calls to flush() and invalidate() are necessary to make changes
        in mapped memory between devices and host available
        """
        ...
    @property
    def memory(self) -> int:
        """
        Mapped memory address of the tensor as seen from the CPU. Zero if not mapped.
        """
        ...
    def retrieve(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Retrieves n elements from the tensor at the given offset and stores it
        in dst.

        Parameters
        ----------
        addr: int
            Address of memory to copy to
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this tensor.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the tensor in bytes.
        """
        ...
    def update(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Updates the tensor at the given offset with n elements stored at addr.

        Parameters
        ----------
        addr: int
            Address of memory to copy from
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...

class FlushMemoryCommand:
    """
    Command for flushing memory writes
    """

    ...

class Geometry:
    """
    Underlying structure Acceleration Structures use to trace rays against
    constructed from Meshes.
    """

    def __init__(self) -> None: ...
    @property
    def blas_address(self) -> int:
        """
        device address of the underlying blas
        """
        ...
    @blas_address.setter
    def blas_address(self, arg: int, /) -> None:
        """
        device address of the underlying blas
        """
        ...
    @property
    def indices_address(self) -> int:
        """
        device address of the index buffer, or zero if it was discarded or is
        non existent
        """
        ...
    @indices_address.setter
    def indices_address(self, arg: int, /) -> None:
        """
        device address of the index buffer, or zero if it was discarded or is
        non existent
        """
        ...
    @property
    def vertices_address(self) -> int:
        """
        device address of the vertex buffer or zero if it was discarded
        """
        ...
    @vertices_address.setter
    def vertices_address(self, arg: int, /) -> None:
        """
        device address of the vertex buffer or zero if it was discarded
        """
        ...

class GeometryInstance:
    """
    Building blocks of Acceleration Structures containing a reference to a
    Geometry along a transformation to be applied to the underlying mesh as
    well as additional information.
    """

    def __init__(self) -> None: ...
    @property
    def blas_address(self) -> int:
        """
        device address of the referenced BLAS/geometry
        """
        ...
    @blas_address.setter
    def blas_address(self, arg: int, /) -> None:
        """
        device address of the referenced BLAS/geometry
        """
        ...
    @property
    def customIndex(self) -> int:
        """
        The custom index of this instance available in the shader.
        """
        ...
    @customIndex.setter
    def customIndex(self, arg: int, /) -> None:
        """
        The custom index of this instance available in the shader.
        """
        ...
    @property
    def mask(self) -> int:
        """
        Mask of this instance used for masking ray traces.
        """
        ...
    @mask.setter
    def mask(self, arg: int, /) -> None:
        """
        Mask of this instance used for masking ray traces.
        """
        ...
    @property
    def transform(self) -> numpy.typing.NDArray:
        """
        The transformation applied to the referenced geometry.
        """
        ...
    @transform.setter
    def transform(self, arg: numpy.typing.NDArray, /) -> None:
        """
        The transformation applied to the referenced geometry.
        """
        ...

class GeometryInstanceVector:
    """
    List of GeometryInstance
    """

    def __init__(
        self, arg: Iterable[hephaistos.pyhephaistos.GeometryInstance], /
    ) -> None:
        """
        Construct from an iterable object
        """
        ...
    @overload
    def __init__(self) -> None:
        """
        Default constructor
        """
        ...
    @overload
    def __init__(self, arg: hephaistos.pyhephaistos.GeometryInstanceVector) -> None:
        """
        Copy constructor
        """
        ...
    def append(self, arg: hephaistos.pyhephaistos.GeometryInstance, /) -> None:
        """
        Append `arg` to the end of the list.
        """
        ...
    def clear(self) -> None:
        """
        Remove all items from list.
        """
        ...
    def extend(self, arg: hephaistos.pyhephaistos.GeometryInstanceVector, /) -> None:
        """
        Extend `self` by appending elements from `arg`.
        """
        ...
    def insert(
        self, arg0: int, arg1: hephaistos.pyhephaistos.GeometryInstance, /
    ) -> None:
        """
        Insert object `arg1` before index `arg0`.
        """
        ...
    def pop(self, index: int = -1) -> hephaistos.pyhephaistos.GeometryInstance:
        """
        Remove and return item at `index` (default last).
        """
        ...

class GeometryStore:
    """
    If True, keeps the mesh data on the GPU after building Geometries.
    """

    def __init__(
        self, meshes: hephaistos.pyhephaistos.MeshVector, keepMeshData: bool = True
    ) -> None:
        """
        Creates a geometry store responsible for managing the BLAS/geometries
        used to create and run acceleration structures.
        """
        ...
    def createInstance(self, idx: int) -> hephaistos.pyhephaistos.GeometryInstance:
        """
        Creates a new instance of the specified geometry
        """
        ...
    @property
    def geometries(self) -> hephaistos.pyhephaistos.GeometryVector:
        """
        Returns the list of stored geometries
        """
        ...
    @property
    def size(self) -> int:
        """
        Number of geometries stored
        """
        ...

class GeometryVector:
    """
    List of Geometry
    """

    def __init__(self, arg: Iterable[hephaistos.pyhephaistos.Geometry], /) -> None:
        """
        Construct from an iterable object
        """
        ...
    @overload
    def __init__(self) -> None:
        """
        Default constructor
        """
        ...
    @overload
    def __init__(self, arg: hephaistos.pyhephaistos.GeometryVector) -> None:
        """
        Copy constructor
        """
        ...
    def append(self, arg: hephaistos.pyhephaistos.Geometry, /) -> None:
        """
        Append `arg` to the end of the list.
        """
        ...
    def clear(self) -> None:
        """
        Remove all items from list.
        """
        ...
    def extend(self, arg: hephaistos.pyhephaistos.GeometryVector, /) -> None:
        """
        Extend `self` by appending elements from `arg`.
        """
        ...
    def insert(self, arg0: int, arg1: hephaistos.pyhephaistos.Geometry, /) -> None:
        """
        Insert object `arg1` before index `arg0`.
        """
        ...
    def pop(self, index: int = -1) -> hephaistos.pyhephaistos.Geometry:
        """
        Remove and return item at `index` (default last).
        """
        ...

class HeaderMap:
    """
    Dict mapping filepaths to shader source code. Consumed by Compiler to
    resolve include directives.
    """

    class ItemView: ...
    class KeyView: ...
    class ValueView: ...

    def __init__(self, arg: dict[str, str], /) -> None:
        """
        Construct from a dictionary
        """
        ...
    @overload
    def __init__(self) -> None:
        """
        Default constructor
        """
        ...
    @overload
    def __init__(self, arg: hephaistos.pyhephaistos.HeaderMap) -> None:
        """
        Copy constructor
        """
        ...
    def clear(self) -> None:
        """
        Remove all items
        """
        ...
    def items(self) -> hephaistos.pyhephaistos.HeaderMap.ItemView:
        """
        Returns an iterable view of the map's items.
        """
        ...
    def keys(self) -> hephaistos.pyhephaistos.HeaderMap.KeyView:
        """
        Returns an iterable view of the map's keys.
        """
        ...
    def update(self, arg: hephaistos.pyhephaistos.HeaderMap, /) -> None:
        """
        Update the map with element from `arg`
        """
        ...
    def values(self) -> hephaistos.pyhephaistos.HeaderMap.ValueView:
        """
        Returns an iterable view of the map's values.
        """
        ...

class Image:
    """
    Allocates memory on the device using a memory layout it deems optimal for
    images presenting it inside programs as storage images thus allowing reads
    and writes to it.

    Parameters
    ----------
    format: ImageFormat
        Format of the image
    width: int
        Width of the image in pixels
    height: int, default=1
        Height of the image in pixels
    depth: int, default=1
        Depth of the image in pixels
    """

    def __init__(
        self,
        format: hephaistos.pyhephaistos.ImageFormat,
        width: int,
        height: int = 1,
        depth: int = 1,
    ) -> None: ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str, /
    ) -> None:
        """
        Binds the image to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int, /
    ) -> None:
        """
        Binds the image to the program at the given binding
        """
        ...
    @property
    def depth(self) -> int:
        """
        Depth of the image in pixels
        """
        ...
    @property
    def format(self) -> hephaistos.pyhephaistos.ImageFormat:
        """
        Format of the image
        """
        ...
    @property
    def height(self) -> int:
        """
        Height of the image in pixels
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        Size the image takes in a linear/compact memory layout in bytes. This
        can differ from the actual size the image takes on the device but can
        be useful to allocate buffers for transferring the image.
        """
        ...
    @property
    def width(self) -> int:
        """
        With of the image in pixels
        """
        ...

class ImageBindingTraits:
    """
    Properties a binding expects from a bound image
    """

    @property
    def dims(self) -> int:
        """
        Image dimensions
        """
        ...
    @property
    def format(self) -> hephaistos.pyhephaistos.ImageFormat:
        """
        Expected image format
        """
        ...

class ImageBuffer:
    """
    Utility class allocating memory on the host side in linear memory layout
    allowing easy manipulating of 2D 32bit RGBA image data that can later be
    copied to an image or texture. Provides additional methods for loading and
    saving image data from and to disk in various file formats.

    Parameters
    ----------
    width: int
        Width of the image in pixels
    height: int
        Height of the image in pixels
    """

    def __init__(self, width: int, height: int) -> None: ...
    @property
    def height(self) -> int:
        """
        Height of the image in pixels
        """
        ...
    def loadFile(filename: str) -> hephaistos.pyhephaistos.ImageBuffer:
        """
        Loads the image at the given filepath and returns a new ImageBuffer
        """
        ...
    def loadMemory(data: bytes) -> hephaistos.pyhephaistos.ImageBuffer:
        """
        Loads serialized image data from memory and returns a new ImageBuffer

        Parameters
        ----------
        data: bytes
            Binary data containing the image data
        """
        ...
    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array that allows to manipulate the data handled
        by this ImageBuffer
        """
        ...
    def save(self, filename: str) -> None:
        """
        Saves the image under the given filepath
        """
        ...
    @property
    def width(self) -> int:
        """
        Width of the image in pixels
        """
        ...

class ImageFormat:
    """
    List of supported image formats
    """

    R16G16B16A16_SINT: ImageFormat

    R16G16B16A16_UINT: ImageFormat

    R32G32B32A32_SFLOAT: ImageFormat

    R32G32B32A32_SINT: ImageFormat

    R32G32B32A32_UINT: ImageFormat

    R32G32_SFLOAT: ImageFormat

    R32G32_SINT: ImageFormat

    R32G32_UINT: ImageFormat

    R32_SFLOAT: ImageFormat

    R32_SINT: ImageFormat

    R32_UINT: ImageFormat

    R8G8B8A8_SINT: ImageFormat

    R8G8B8A8_SNORM: ImageFormat

    R8G8B8A8_UINT: ImageFormat

    R8G8B8A8_UNORM: ImageFormat

class IntBuffer:
    """
    Buffer representing memory allocated on the host holding an array of type
    int32 and given size which can be accessed as numpy array.

    Parameters
    ----------
    size: int
        Number of elements
    """

    def __init__(self, size: int) -> None: ...
    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array using this buffer's memory.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this buffer.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class IntTensor:
    """
    Tensor representing memory allocated on the device holding an array of type
    int32 and given size. If mapped can be accessed from the host using its
    memory address. Mapping can optionally be requested but will be ignored if
    the device does not support it. Query its support after creation via isMapped.
    """

    def __init__(self, array: numpy.typing.NDArray, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        array: NDArray
            Numpy array containing the data used to fill the tensor.
            Its type must match the tensor's.
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, size: int, mapped: bool = False) -> None:
        """
        Creates a new tensor of given size.

        Parameters
        ----------
        size: int
            Number of elements
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, addr: int, n: int, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        addr: int
            Address of the data used to fill the tensor
        n: int
            Number of bytes to copy from addr
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @property
    def address(self) -> int:
        """
        The device address of this tensor.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    def flush(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the host available to the device.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to flush
        size: int|None, default=None
            Number of elements to flush starting at offset.
            If None, flushes all remaining elements.
        """
        ...
    def invalidate(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the device available to the host.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to invalidate
        size: int | None, default=None
            Number of elements to invalidate starting at offset.
            If None, invalidates all remaining elements.
        """
        ...
    @property
    def isMapped(self) -> bool:
        """
        True, if the underlying memory is writable by the CPU.
        """
        ...
    @property
    def isNonCoherent(self) -> bool:
        """
        Wether calls to flush() and invalidate() are necessary to make changes
        in mapped memory between devices and host available
        """
        ...
    @property
    def memory(self) -> int:
        """
        Mapped memory address of the tensor as seen from the CPU. Zero if not mapped.
        """
        ...
    def retrieve(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Retrieves n elements from the tensor at the given offset and stores it
        in dst.

        Parameters
        ----------
        addr: int
            Address of memory to copy to
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this tensor.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the tensor in bytes.
        """
        ...
    def update(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Updates the tensor at the given offset with n elements stored at addr.

        Parameters
        ----------
        addr: int
            Address of memory to copy from
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...

class LocalSize:
    """
    Description if the local size, i.e. the number and arrangement of threads in
    a single thread group
    """

    def __init__(self) -> None: ...
    @property
    def x(self) -> int:
        """
        Number of threads in X dimension
        """
        ...
    @x.setter
    def x(self, arg: int, /) -> None:
        """
        Number of threads in X dimension
        """
        ...
    @property
    def y(self) -> int:
        """
        Number of threads in Y dimension
        """
        ...
    @y.setter
    def y(self, arg: int, /) -> None:
        """
        Number of threads in Y dimension
        """
        ...
    @property
    def z(self) -> int:
        """
        Number of threads in Z dimension
        """
        ...
    @z.setter
    def z(self, arg: int, /) -> None:
        """
        Number of threads in Z dimension
        """
        ...

class LongBuffer:
    """
    Buffer representing memory allocated on the host holding an array of type
    int64 and given size which can be accessed as numpy array.

    Parameters
    ----------
    size: int
        Number of elements
    """

    def __init__(self, size: int) -> None: ...
    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array using this buffer's memory.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this buffer.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class LongTensor:
    """
    Tensor representing memory allocated on the device holding an array of type
    int64 and given size. If mapped can be accessed from the host using its
    memory address. Mapping can optionally be requested but will be ignored if
    the device does not support it. Query its support after creation via isMapped.
    """

    def __init__(self, array: numpy.typing.NDArray, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        array: NDArray
            Numpy array containing the data used to fill the tensor.
            Its type must match the tensor's.
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, size: int, mapped: bool = False) -> None:
        """
        Creates a new tensor of given size.

        Parameters
        ----------
        size: int
            Number of elements
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, addr: int, n: int, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        addr: int
            Address of the data used to fill the tensor
        n: int
            Number of bytes to copy from addr
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @property
    def address(self) -> int:
        """
        The device address of this tensor.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    def flush(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the host available to the device.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to flush
        size: int|None, default=None
            Number of elements to flush starting at offset.
            If None, flushes all remaining elements.
        """
        ...
    def invalidate(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the device available to the host.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to invalidate
        size: int | None, default=None
            Number of elements to invalidate starting at offset.
            If None, invalidates all remaining elements.
        """
        ...
    @property
    def isMapped(self) -> bool:
        """
        True, if the underlying memory is writable by the CPU.
        """
        ...
    @property
    def isNonCoherent(self) -> bool:
        """
        Wether calls to flush() and invalidate() are necessary to make changes
        in mapped memory between devices and host available
        """
        ...
    @property
    def memory(self) -> int:
        """
        Mapped memory address of the tensor as seen from the CPU. Zero if not mapped.
        """
        ...
    def retrieve(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Retrieves n elements from the tensor at the given offset and stores it
        in dst.

        Parameters
        ----------
        addr: int
            Address of memory to copy to
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this tensor.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the tensor in bytes.
        """
        ...
    def update(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Updates the tensor at the given offset with n elements stored at addr.

        Parameters
        ----------
        addr: int
            Address of memory to copy from
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...

class Mesh:
    """
    Representation of a geometric shape consisting of triangles defined by their
    vertices, which can optionally be indexed. Meshes are used to build Geometries.
    """

    def __init__(self) -> None: ...
    @property
    def indices(self) -> numpy.typing.NDArray:
        """
        Optional numpy array holding the indices referencing vertices to create
        triangles.
        """
        ...
    @indices.setter
    def indices(self, arg: numpy.typing.NDArray, /) -> None:
        """
        Optional numpy array holding the indices referencing vertices to create
        triangles.
        """
        ...
    @property
    def vertices(self) -> numpy.typing.NDArray:
        """
        Numpy array holding the vertex data. The first three columns must be the
        x, y and z positions.
        """
        ...
    @vertices.setter
    def vertices(self, arg: numpy.typing.NDArray, /) -> None:
        """
        Numpy array holding the vertex data. The first three columns must be the
        x, y and z positions.
        """
        ...

class MeshVector:
    """
    List of Mesh
    """

    def __init__(self, arg: Iterable[hephaistos.pyhephaistos.Mesh], /) -> None:
        """
        Construct from an iterable object
        """
        ...
    @overload
    def __init__(self) -> None:
        """
        Default constructor
        """
        ...
    @overload
    def __init__(self, arg: hephaistos.pyhephaistos.MeshVector) -> None:
        """
        Copy constructor
        """
        ...
    def append(self, arg: hephaistos.pyhephaistos.Mesh, /) -> None:
        """
        Append `arg` to the end of the list.
        """
        ...
    def clear(self) -> None:
        """
        Remove all items from list.
        """
        ...
    def extend(self, arg: hephaistos.pyhephaistos.MeshVector, /) -> None:
        """
        Extend `self` by appending elements from `arg`.
        """
        ...
    def insert(self, arg0: int, arg1: hephaistos.pyhephaistos.Mesh, /) -> None:
        """
        Insert object `arg1` before index `arg0`.
        """
        ...
    def pop(self, index: int = -1) -> hephaistos.pyhephaistos.Mesh:
        """
        Remove and return item at `index` (default last).
        """
        ...

class ParameterType:
    """
    Type of parameter
    """

    ACCELERATION_STRUCTURE: ParameterType

    COMBINED_IMAGE_SAMPLER: ParameterType

    STORAGE_BUFFER: ParameterType

    STORAGE_IMAGE: ParameterType

    UNIFORM_BUFFER: ParameterType

class Program:
    """
    Encapsulates a shader program enabling introspection into its bindings as
    well as keeping track of the parameters currently bound to them. Execution
    happens trough commands.
    """

    def __init__(self, code: bytes, specialization: bytes) -> None:
        """
        Creates a new program using the shader's byte code.

        Parameters
        ----------
        code: bytes
            Byte code of the program
        specialization: bytes
            Data used for filling in specialization constants
        """
        ...
    @overload
    def __init__(self, code: bytes) -> None:
        """
        Creates a new program using the shader's byte code.

        Parameters
        ----------
        code: bytes
            Byte code of the program
        """
        ...
    def bindParams(*params, **namedparams) -> None:
        """
        Binds the given parameters.
        Positional arguments are bound to the binding of the corresponding
        position. Keyword arguments are matched with the binding of the same
        name.
        """
        ...
    @property
    def bindings(self) -> list[hephaistos.pyhephaistos.BindingTraits]:
        """
        Returns a list of all bindings.
        """
        ...
    def dispatch(
        self, x: int = 1, y: int = 1, z: int = 1
    ) -> hephaistos.pyhephaistos.DispatchCommand:
        """
        Dispatches a program execution with the given amount of workgroups.

        Parameters
        ----------
        x: int, default=1
            Number of groups to dispatch in X dimension
        y: int, default=1
            Number of groups to dispatch in Y dimension
        z: int, default=1
            Number of groups to dispatch in Z dimension
        """
        ...
    def dispatchIndirect(
        self, tensor: hephaistos.pyhephaistos.Tensor, offset: int = 0
    ) -> hephaistos.pyhephaistos.DispatchIndirectCommand:
        """
        Dispatches a program execution using the amount of workgroups stored in
        the given tensor at the given offset. Expects the workgroup size as
        three consecutive unsigned 32 bit integers.

        Parameters
        ----------
        tensor: Tensor
            Tensor from which to read the amount of workgroups
        offset: int, default=0
            Offset at which to start reading
        """
        ...
    def dispatchIndirectPush(
        self, push: bytes, tensor: hephaistos.pyhephaistos.Tensor, offset: int = 0
    ) -> hephaistos.pyhephaistos.DispatchIndirectCommand:
        """
        Dispatches a program execution with the given push data using the amount
        of workgroups stored in the given tensor at the given offset.

        Parameters
        ----------
        push: bytes
            Data pushed to the dispatch as bytes
        tensor: Tensor
            Tensor from which to read the amount of workgroups
        offset: int, default=0
            Offset at which to start reading
        """
        ...
    def dispatchPush(
        self, push: bytes, x: int = 1, y: int = 1, z: int = 1
    ) -> hephaistos.pyhephaistos.DispatchCommand:
        """
        Dispatches a program execution with the given push data and amount of
        workgroups.

        Parameters
        ----------
        push: bytes
            Data pushed to the dispatch as bytes
        x: int, default=1
            Number of groups to dispatch in X dimension
        y: int, default=1
            Number of groups to dispatch in Y dimension
        z: int, default=1
            Number of groups to dispatch in Z dimension
        """
        ...
    def isBindingBound(i: int,/) -> bool:
        """
        Checks wether the i-th binding is currently bound.
        """
        ...
    @overload
    def isBindingBound(name: str,/) -> bool:
        """
        Checks wether the binding specified by its name is currently bound
        """
        ...
    @property
    def localSize(self) -> hephaistos.pyhephaistos.LocalSize:
        """
        Returns the size of the local work group.
        """
        ...

R16G16B16A16_SINT: ImageFormat

R16G16B16A16_UINT: ImageFormat

R32G32B32A32_SFLOAT: ImageFormat

R32G32B32A32_SINT: ImageFormat

R32G32B32A32_UINT: ImageFormat

R32G32_SFLOAT: ImageFormat

R32G32_SINT: ImageFormat

R32G32_UINT: ImageFormat

R32_SFLOAT: ImageFormat

R32_SINT: ImageFormat

R32_UINT: ImageFormat

R8G8B8A8_SINT: ImageFormat

R8G8B8A8_SNORM: ImageFormat

R8G8B8A8_UINT: ImageFormat

R8G8B8A8_UNORM: ImageFormat

class RawBuffer:
    """
    Buffer for allocating a raw chunk of memory on the host accessible via its
    memory address. Useful as a base class providing more complex functionality.

    Parameters
    ----------
    size: int
        size of the buffer in bytes
    """

    def __init__(self, size: int) -> None: ...
    @property
    def address(self) -> int:
        """
        The memory address of the allocated buffer.
        """
        ...
    @property
    def size(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class RetrieveImageCommand:
    """
    Command for copying the image back into the given buffer
    """

    def __init__(
        self, src: hephaistos.pyhephaistos.Image, dst: hephaistos.pyhephaistos.Buffer, /
    ) -> None: ...

class RetrieveTensorCommand:
    """
    Command for copying the src tensor back to the destination buffer
    """

    def __init__(
        self,
        src: hephaistos.pyhephaistos.Tensor,
        dst: hephaistos.pyhephaistos.Buffer,
        bufferOffset: Optional[int] = None,
        tensorOffset: Optional[int] = None,
        size: Optional[int] = None,
        unsafe: bool = False
    ) -> None: ...

class SequenceBuilder:
    """
    Builder class for recording a sequence of commands and subroutines and
    submitting it to the device for execution, which happens asynchronous, i.e.
    submit() returns before the recorded work finishes.
    """

    def And(
        self, subroutine: hephaistos.pyhephaistos.Subroutine
    ) -> hephaistos.pyhephaistos.SequenceBuilder:
        """
        Issues the subroutine to run parallel in the current step.
        """
        ...
    @overload
    def And(
        self, cmd: hephaistos.pyhephaistos.Command
    ) -> hephaistos.pyhephaistos.SequenceBuilder:
        """
        Issues the command to run parallel in the current step.
        """
        ...
    def AndList(self, list: list) -> hephaistos.pyhephaistos.SequenceBuilder:
        """
        Issues each element of the list to run parallel in the current step
        """
        ...
    def NextStep(self) -> hephaistos.pyhephaistos.SequenceBuilder:
        """
        Issues a new step. Following calls to And are ensured to run after previous ones finished.
        """
        ...
    def Submit(self) -> hephaistos.pyhephaistos.Submission:
        """
        Submits the recorded steps as a single batch to the GPU.
        """
        ...
    def Then(
        self, subroutine: hephaistos.pyhephaistos.Subroutine
    ) -> hephaistos.pyhephaistos.SequenceBuilder:
        """
        Issues a new step to execute after waiting for the previous one to finish.
        """
        ...
    @overload
    def Then(
        self, cmd: hephaistos.pyhephaistos.Command
    ) -> hephaistos.pyhephaistos.SequenceBuilder:
        """
        Issues a new step to execute after waiting for the previous one to finish.
        """
        ...
    def WaitFor(
        self, timeline: hephaistos.pyhephaistos.Timeline, value: int
    ) -> hephaistos.pyhephaistos.SequenceBuilder:
        """
        Issues the following steps to wait for the given timeline to reach the given value
        """
        ...
    @overload
    def WaitFor(self, value: int) -> hephaistos.pyhephaistos.SequenceBuilder:
        """
        Issues the following steps to wait on the sequence timeline to reach the given value.
        """
        ...
    def __init__(
        self, timeline: hephaistos.pyhephaistos.Timeline, startValue: int
    ) -> None:
        """
        Creates a new SequenceBuilder.

        Parameters
        ----------
        timeline: Timeline
            Timeline to use for orchestrating commands and subroutines
        startValue: int
            Counter value to wait for on the timeline to start with
        """
        ...
    @overload
    def __init__(self, timeline: hephaistos.pyhephaistos.Timeline) -> None:
        """
        Creates a new SequenceBuilder.

        Parameters
        ----------
        timeline: Timeline
            Timeline to use for orchestrating commands and subroutines
        """
        ...
    def printWaitGraph(self) -> str:
        """
        Returns a visualization of the current wait graph in the form:
        (Timeline.ID(WaitValue))* -> (submissions) -> (Timeline.ID(SignalValue)).
        Must be called before Submit().
        """
        ...

class ShortBuffer:
    """
    Buffer representing memory allocated on the host holding an array of type
    int16 and given size which can be accessed as numpy array.

    Parameters
    ----------
    size: int
        Number of elements
    """

    def __init__(self, size: int) -> None: ...
    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array using this buffer's memory.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this buffer.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class ShortTensor:
    """
    Tensor representing memory allocated on the device holding an array of type
    int16 and given size. If mapped can be accessed from the host using its
    memory address. Mapping can optionally be requested but will be ignored if
    the device does not support it. Query its support after creation via isMapped.
    """

    def __init__(self, array: numpy.typing.NDArray, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        array: NDArray
            Numpy array containing the data used to fill the tensor.
            Its type must match the tensor's.
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, size: int, mapped: bool = False) -> None:
        """
        Creates a new tensor of given size.

        Parameters
        ----------
        size: int
            Number of elements
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, addr: int, n: int, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        addr: int
            Address of the data used to fill the tensor
        n: int
            Number of bytes to copy from addr
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @property
    def address(self) -> int:
        """
        The device address of this tensor.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    def flush(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the host available to the device.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to flush
        size: int|None, default=None
            Number of elements to flush starting at offset.
            If None, flushes all remaining elements.
        """
        ...
    def invalidate(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the device available to the host.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to invalidate
        size: int | None, default=None
            Number of elements to invalidate starting at offset.
            If None, invalidates all remaining elements.
        """
        ...
    @property
    def isMapped(self) -> bool:
        """
        True, if the underlying memory is writable by the CPU.
        """
        ...
    @property
    def isNonCoherent(self) -> bool:
        """
        Wether calls to flush() and invalidate() are necessary to make changes
        in mapped memory between devices and host available
        """
        ...
    @property
    def memory(self) -> int:
        """
        Mapped memory address of the tensor as seen from the CPU. Zero if not mapped.
        """
        ...
    def retrieve(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Retrieves n elements from the tensor at the given offset and stores it
        in dst.

        Parameters
        ----------
        addr: int
            Address of memory to copy to
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this tensor.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the tensor in bytes.
        """
        ...
    def update(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Updates the tensor at the given offset with n elements stored at addr.

        Parameters
        ----------
        addr: int
            Address of memory to copy from
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...

class StopWatch:
    """
    Allows the measuring of elapsed time between commands execution
    
    """

    def __init__(self) -> None:
        """
        Creates a new stopwatch for measuring elapsed time between commands.
        """
        ...
    def getElapsedTime(self, wait: bool = False) -> list[float]:
        """
        Retrieves the elapsed time between start() and stop() in nanoseconds.
        If wait = True, blocks the caller until all timestamps are available,
        otherwise returns NaN.
        """
        ...
    def reset(self) -> None:
        """
        Resets the stop watch.
        """
        ...
    def start(self) -> hephaistos.pyhephaistos.Command:
        """
        Returns the command to start the stop watch.
        """
        ...
    def stop(self) -> hephaistos.pyhephaistos.Command:
        """
        Returns the command to stop the stop watch. Can be recorded multiple
        times to record up to stopCount stop times.
        """
        ...

class SubgroupProperties:
    """
    List of subgroup properties and supported operations
    """

    @property
    def arithmeticSupport(self) -> bool:
        """
        Support for GL_KHR_shader_subgroup_arithmetic
        """
        ...
    @property
    def ballotSupport(self) -> bool:
        """
        Support for GL_KHR_shader_subgroup_ballot
        """
        ...
    @property
    def basicSupport(self) -> bool:
        """
        Support for GL_KHR_shader_subgroup_basic
        """
        ...
    @property
    def quadSupport(self) -> bool:
        """
        Support for GL_KHR_shader_subgroup_quad
        """
        ...
    @property
    def shuffleClusteredSupport(self) -> bool:
        """
        Support for GL_KHR_shader_subgroup_clustered
        """
        ...
    @property
    def shuffleRelativeSupport(self) -> bool:
        """
        Support for GL_KHR_shader_subgroup_shuffle_relative
        """
        ...
    @property
    def shuffleSupport(self) -> bool:
        """
        Support for GL_KHR_shader_subgroup_shuffle
        """
        ...
    @property
    def subgroupSize(self) -> int:
        """
        Threads per subgroup
        """
        ...
    @property
    def voteSupport(self) -> bool:
        """
        Support for GL_KHR_shader_subgroup_vote
        """
        ...

class Submission:
    """
    Submissions are issued after work has been submitted to the device and can
    be used to wait for its completion.
    """

    @property
    def finalStep(self) -> int:
        """
        The value the timeline will reach when the submission finishes.
        """
        ...
    @property
    def forgettable(self) -> bool:
        """
        True, if the Submission can be discarded safely, i.e. fire and forget.
        """
        ...
    @property
    def timeline(self) -> hephaistos.pyhephaistos.Timeline:
        """
        The timeline this submission was issued with.
        """
        ...
    def wait(self) -> None:
        """
        Blocks the caller until the submission finishes.
        """
        ...
    def waitTimeout(self, ns: int) -> bool:
        """
        Blocks the caller until the submission finished or the specified time
        elapsed. Returns True in the first, False in the second case.
        """
        ...

class Subroutine:
    """
    Subroutines are reusable sequences of commands that can be submitted
    multiple times to the device. Recording sequence of commands require non
    negligible CPU time and may be amortized by reusing sequences via Subroutines.
    """

    @property
    def simultaneousUse(self) -> bool:
        """
        True, if the subroutine can be used simultaneous
        """
        ...

class Tensor:
    """
    Base class for all tensors managing memory allocations on the device
    """

    ...

class Texture:
    """
    Allocates memory on the device using a memory layout it deems optimal for
    images and presents it inside programs as texture allowing filtered lookups.
    The filter methods can specified

    Parameters
    ----------
    format: ImageFormat
        Format of the image
    width: int
        Width of the image in pixels
    height: int, default=1
        Height of the image in pixels
    depth: int, default=1
        Depth of the image in pixels
    filter: 'nearest'|'n'|'linear'|'l', default='linear'
        Method used to interpolate between pixels
    unnormalized: bool, default=False
        If True, use coordinates in pixel space rather than normalized ones
        inside programs
    modeU: 'repeat'|'r'|'mirrored repeat'|'mr'|'clamp edge'|'ce'|'mirrored clamp edge'|'mce', default='repeat'
        Method used to handle out of range coordinates in U dimension
    modeV: 'repeat'|'r'|'mirrored repeat'|'mr'|'clamp edge'|'ce'|'mirrored clamp edge'|'mce', default='repeat'
        Method used to handle out of range coordinates in V dimension
    modeW: 'repeat'|'r'|'mirrored repeat'|'mr'|'clamp edge'|'ce'|'mirrored clamp edge'|'mce', default='repeat'
        Method used to handle out of range coordinates in W dimension
    """

    def __init__(
        format: hephaistos.pyhephaistos.ImageFormat,
        width: int,
        height: int = 1,
        depth: int = 1,
        *,
        filter: Literal["nearest", "n", "linear", "l"] = "linear",
        unnormalized: bool = False,
        modeU: Literal[
            "repeat",
            "r",
            "mirrored repeat",
            "mr",
            "clamp edge",
            "ce",
            "mirrored clamp edge",
            "mce",
        ] = "repeat",
        modeV: Literal[
            "repeat",
            "r",
            "mirrored repeat",
            "mr",
            "clamp edge",
            "ce",
            "mirrored clamp edge",
            "mce",
        ] = "repeat",
        modeW: Literal[
            "repeat",
            "r",
            "mirrored repeat",
            "mr",
            "clamp edge",
            "ce",
            "mirrored clamp edge",
            "mce",
        ] = "repeat",
    ) -> None: ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str, /
    ) -> None:
        """
        Binds the texture to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int, /
    ) -> None:
        """
        Binds the texture to the program at the given binding
        """
        ...
    @property
    def depth(self) -> int:
        """
        Depth of the texture in pixels
        """
        ...
    @property
    def format(self) -> hephaistos.pyhephaistos.ImageFormat:
        """
        Format of the texture
        """
        ...
    @property
    def height(self) -> int:
        """
        Height of the texture in pixels
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        Size the texture takes in a linear/compact memory layout in bytes. This
        can differ from the actual size the texture takes on the device but can
        be useful to allocate buffers for transferring the texture.
        """
        ...
    @property
    def width(self) -> int:
        """
        Width of the texture in pixels
        """
        ...

class Timeline:
    """
    Timeline managing the execution of code and commands using an increasing
    internal counter. Both GPU and CPU can wait and/or increase this counter
    thus creating a synchronization between them or among themselves. The
    current value of the counter can be queried allowing an asynchronous method
    of reporting the progress.

    Parameters
    ----------
    value: int, default=0
        Initial value of the internal counter
    """

    def __init__(self, initialValue: int) -> None: ...
    @overload
    def __init__(self) -> None: ...
    @property
    def id(self) -> int:
        """
        Id of this timeline
        """
        ...
    @property
    def value(self) -> int:
        """
        Returns or sets the current value of the timeline. Note that the value
        can only increase. Disobeying this requirement results in undefined
        behavior.
        """
        ...
    @value.setter
    def value(self, arg: int, /) -> None:
        """
        Returns or sets the current value of the timeline. Note that the value
        can only increase. Disobeying this requirement results in undefined
        behavior.
        """
        ...
    def wait(self, value: int) -> None:
        """
        Waits for the timeline to reach the given value.
        """
        ...
    def waitTimeout(self, value: int, timeout: int) -> bool:
        """
        Waits for the timeline to reach the given value for a certain amount.
        Returns True if the value was reached and False if it timed out.

        Parameters
        ----------
        value: int
            Value to wait for
        timeout: int
            Time in nanoseconds to wait. May be rounded to the closest internal
            precision of the device clock.
        """
        ...


class TypeSupport:
    """List of supported extended types, e.g. float64"""

    @property
    def float64(self) -> bool: ...
    @property
    def float16(self) -> bool: ...
    @property
    def int64(self) -> bool: ...
    @property
    def int16(self) -> bool: ...
    @property
    def int8(self) -> bool: ...


class UnsignedIntBuffer:
    """
    Buffer representing memory allocated on the host holding an array of type
    uint32 and given size which can be accessed as numpy array.

    Parameters
    ----------
    size: int
        Number of elements
    """

    def __init__(self, size: int) -> None: ...
    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array using this buffer's memory.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this buffer.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class UnsignedIntTensor:
    """
    Tensor representing memory allocated on the device holding an array of type
    uint32 and given size. If mapped can be accessed from the host using its
    memory address. Mapping can optionally be requested but will be ignored if
    the device does not support it. Query its support after creation via isMapped.
    """

    def __init__(self, array: numpy.typing.NDArray, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        array: NDArray
            Numpy array containing the data used to fill the tensor.
            Its type must match the tensor's.
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, size: int, mapped: bool = False) -> None:
        """
        Creates a new tensor of given size.

        Parameters
        ----------
        size: int
            Number of elements
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, addr: int, n: int, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        addr: int
            Address of the data used to fill the tensor
        n: int
            Number of bytes to copy from addr
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @property
    def address(self) -> int:
        """
        The device address of this tensor.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    def flush(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the host available to the device.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to flush
        size: int|None, default=None
            Number of elements to flush starting at offset.
            If None, flushes all remaining elements.
        """
        ...
    def invalidate(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the device available to the host.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to invalidate
        size: int | None, default=None
            Number of elements to invalidate starting at offset.
            If None, invalidates all remaining elements.
        """
        ...
    @property
    def isMapped(self) -> bool:
        """
        True, if the underlying memory is writable by the CPU.
        """
        ...
    @property
    def isNonCoherent(self) -> bool:
        """
        Wether calls to flush() and invalidate() are necessary to make changes
        in mapped memory between devices and host available
        """
        ...
    @property
    def memory(self) -> int:
        """
        Mapped memory address of the tensor as seen from the CPU. Zero if not mapped.
        """
        ...
    def retrieve(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Retrieves n elements from the tensor at the given offset and stores it
        in dst.

        Parameters
        ----------
        addr: int
            Address of memory to copy to
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this tensor.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the tensor in bytes.
        """
        ...
    def update(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Updates the tensor at the given offset with n elements stored at addr.

        Parameters
        ----------
        addr: int
            Address of memory to copy from
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...

class UnsignedLongBuffer:
    """
    Buffer representing memory allocated on the host holding an array of type
    uint64 and given size which can be accessed as numpy array.

    Parameters
    ----------
    size: int
        Number of elements
    """

    def __init__(self, size: int) -> None: ...
    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array using this buffer's memory.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this buffer.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class UnsignedLongTensor:
    """
    Tensor representing memory allocated on the device holding an array of type
    uint64 and given size. If mapped can be accessed from the host using its
    memory address. Mapping can optionally be requested but will be ignored if
    the device does not support it. Query its support after creation via isMapped.
    """

    def __init__(self, array: numpy.typing.NDArray, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        array: NDArray
            Numpy array containing the data used to fill the tensor.
            Its type must match the tensor's.
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, size: int, mapped: bool = False) -> None:
        """
        Creates a new tensor of given size.

        Parameters
        ----------
        size: int
            Number of elements
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, addr: int, n: int, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        addr: int
            Address of the data used to fill the tensor
        n: int
            Number of bytes to copy from addr
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @property
    def address(self) -> int:
        """
        The device address of this tensor.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    def flush(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the host available to the device.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to flush
        size: int|None, default=None
            Number of elements to flush starting at offset.
            If None, flushes all remaining elements.
        """
        ...
    def invalidate(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the device available to the host.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to invalidate
        size: int | None, default=None
            Number of elements to invalidate starting at offset.
            If None, invalidates all remaining elements.
        """
        ...
    @property
    def isMapped(self) -> bool:
        """
        True, if the underlying memory is writable by the CPU.
        """
        ...
    @property
    def isNonCoherent(self) -> bool:
        """
        Wether calls to flush() and invalidate() are necessary to make changes
        in mapped memory between devices and host available
        """
        ...
    @property
    def memory(self) -> int:
        """
        Mapped memory address of the tensor as seen from the CPU. Zero if not mapped.
        """
        ...
    def retrieve(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Retrieves n elements from the tensor at the given offset and stores it
        in dst.

        Parameters
        ----------
        addr: int
            Address of memory to copy to
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this tensor.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the tensor in bytes.
        """
        ...
    def update(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Updates the tensor at the given offset with n elements stored at addr.

        Parameters
        ----------
        addr: int
            Address of memory to copy from
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...

class UnsignedShortBuffer:
    """
    Buffer representing memory allocated on the host holding an array of type
    uint16 and given size which can be accessed as numpy array.

    Parameters
    ----------
    size: int
        Number of elements
    """

    def __init__(self, size: int) -> None: ...
    def numpy(self) -> numpy.typing.NDArray:
        """
        Returns a numpy array using this buffer's memory.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this buffer.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the buffer in bytes.
        """
        ...

class UnsignedShortTensor:
    """
    Tensor representing memory allocated on the device holding an array of type
    uint16 and given size. If mapped can be accessed from the host using its
    memory address. Mapping can optionally be requested but will be ignored if
    the device does not support it. Query its support after creation via isMapped.
    """

    def __init__(self, array: numpy.typing.NDArray, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        array: NDArray
            Numpy array containing the data used to fill the tensor.
            Its type must match the tensor's.
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, size: int, mapped: bool = False) -> None:
        """
        Creates a new tensor of given size.

        Parameters
        ----------
        size: int
            Number of elements
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @overload
    def __init__(self, addr: int, n: int, mapped: bool = False) -> None:
        """
        Creates a new tensor and fills it with the provided data

        Parameters
        ----------
        addr: int
            Address of the data used to fill the tensor
        n: int
            Number of bytes to copy from addr
        mapped: bool, default=False
            If True, tries to map memory to host address space
        """
        ...
    @property
    def address(self) -> int:
        """
        The device address of this tensor.
        """
        ...
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: str
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    @overload
    def bindParameter(
        self, program: hephaistos.pyhephaistos.Program, binding: int
    ) -> None:
        """
        Binds the tensor to the program at the given binding
        """
        ...
    def flush(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the host available to the device.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to flush
        size: int|None, default=None
            Number of elements to flush starting at offset.
            If None, flushes all remaining elements.
        """
        ...
    def invalidate(self, offset: int = 0, size: int | None = None, /) -> None:
        """
        Makes writes in mapped memory from the device available to the host.
        Only needed if isNonCoherent is True.

        Parameters
        ----------
        offset: int, default=0
            Offset in amount of elements into mapped memory to invalidate
        size: int | None, default=None
            Number of elements to invalidate starting at offset.
            If None, invalidates all remaining elements.
        """
        ...
    @property
    def isMapped(self) -> bool:
        """
        True, if the underlying memory is writable by the CPU.
        """
        ...
    @property
    def isNonCoherent(self) -> bool:
        """
        Wether calls to flush() and invalidate() are necessary to make changes
        in mapped memory between devices and host available
        """
        ...
    @property
    def memory(self) -> int:
        """
        Mapped memory address of the tensor as seen from the CPU. Zero if not mapped.
        """
        ...
    def retrieve(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Retrieves n elements from the tensor at the given offset and stores it
        in dst.

        Parameters
        ----------
        addr: int
            Address of memory to copy to
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...
    @property
    def size(self) -> int:
        """
        The number of elements in this tensor.
        """
        ...
    @property
    def size_bytes(self) -> int:
        """
        The size of the tensor in bytes.
        """
        ...
    def update(self, addr: int, n: int, offset: int = 0, /) -> None:
        """
        Updates the tensor at the given offset with n elements stored at addr.

        Parameters
        ----------
        addr: int
            Address of memory to copy from
        n: int
            Amount of elements to copy
        offset: int, default=0
            Offset into the tensor in amount of elements where the copy starts
        
        Note
        ----
        This operation is only guaranteed to succeed on mapped tensors.
        """
        ...

class UpdateImageCommand:
    """
    Command for copying data from the given buffer into the image
    """

    def __init__(
        self,
        src: hephaistos.pyhephaistos.Buffer,
        dst: hephaistos.pyhephaistos.Image,
        /,
    ) -> None: ...

class UpdateTensorCommand:
    """
    Command for copying the src buffer into the destination tensor
    """

    def __init__(
        self,
        src: hephaistos.pyhephaistos.Buffer,
        dst: hephaistos.pyhephaistos.Tensor,
        bufferOffset: Optional[int] = None,
        tensorOffset: Optional[int] = None,
        size: Optional[int] = None,
        unsafe: bool = False,
    ) -> None: ...

class UpdateTextureCommand:
    """
    Command for copying data from the given buffer into the texture
    """

    def __init__(
        self,
        src: hephaistos.pyhephaistos.Buffer,
        dst: hephaistos.pyhephaistos.Texture,
        /,
    ) -> None: ...

def beginSequence() -> hephaistos.pyhephaistos.SequenceBuilder:
    """
    Starts a new sequence.
    """
    ...

@overload
def beginSequence(
    timeline: hephaistos.pyhephaistos.Timeline, startValue: int = 0
) -> hephaistos.pyhephaistos.SequenceBuilder:
    """
    Starts a new sequence.
    """
    ...

def clearTensor(
    tensor: hephaistos.pyhephaistos.Tensor,
    offset: Optional[int] = None,
    size: Optional[int] = None,
    data: Optional[int] = None,
    unsafe: bool = False,
) -> hephaistos.pyhephaistos.ClearTensorCommand:
    """
    Creates a command for filling a tensor with constant data over a given range.
    Defaults to zeroing the complete tensor

    Parameters
    ----------
    tensor: Tensor
        Tensor to be modified
    offset: None|int, default=None
        Offset into the Tensor at which to start clearing it.
        Defaults to the start of the Tensor.
    size: None|int, default=None
        Amount of bytes to clear. If None, equals to the range starting at
        offset until the end of the tensor
    data: None|int, default=None
        32 bit integer used to fill the tensor. If None, uses all zeros.
    unsafe: bool, default=False
        Wether to omit barrier ensuring read after write ordering.
    """
    ...

def configureDebug(
    enablePrint: bool = False,
    enableGPUValidation: bool = False,
    enableSynchronizationValidation: bool = False,
    enableThreadSafetyValidation: bool = False,
    enableAPIValidation: bool = False,
    callback: Callable[[hephaistos.pyhephaistos.DebugMessage], None] = lambda msg: None,
) -> None:
    """
    Configures the current debug state. Must be called before any other library
    calls except isVulkanAvailable() and isDebugAvailable().

    Parameters
    ----------
    enablePrint: bool, default=False
        Enables the usage of GL_EXT_debug_printf
    enableGPUValidation: bool, default=False
        Enable GPU assisted validation
    enableSynchronizationValidation: bool, default=False
        Enable synchronization validation between resources
    enableThreadSafetyValidation: bool, default=False
        Enable thread safety validation
    enableAPIValidation: bool, default=False
        Enables validation of the Vulkan API usage
    callback: (DebugMessage) -> None, default=print(message)
        Callback called for each message. Prints the message by default.

    Note
    ----
    Debugging requires the Vulkan Validation Layers to be installed.
    """
    ...

def createSubroutine(
    commands: list, simultaneous: bool = False
) -> hephaistos.pyhephaistos.Subroutine:
    """
    creates a subroutine from the list of commands

    Parameters
    ----------
    commands: Command[]
        Sequence of commands the Subroutine consists of
    simultaneous: bool, default=False
        True, if the subroutine can be submitted while a previous submission
        has not yet finished. Disobeying this requirement results in undefined
        behavior.
    """
    ...

def enableAtomics(flags: set, force: bool = False) -> None:
    """
    Enables the atomic features contained in the given set by their name. Set
    force=True if an existing context should be destroyed.
    """
    ...

def enableRaytracing(force: bool = False) -> None:
    """
    Enables ray tracing. (Lazy) context creation fails if not supported. Set
    force=True if an existing context should be destroyed.
    """
    ...

def enumerateDevices() -> list[hephaistos.pyhephaistos.Device]:
    """
    Returns a list of all supported installed devices.
    """
    ...

def execute(sub: hephaistos.pyhephaistos.Subroutine) -> None:
    """
    Runs the given subroutine synchronous.
    """
    ...

@overload
def execute(cmd: hephaistos.pyhephaistos.Command) -> None:
    """
    Runs the given command synchronous.
    """
    ...

def executeList(list: list) -> None:
    """
    Runs the given of list commands synchronous
    """
    ...

def flushMemory() -> hephaistos.pyhephaistos.FlushMemoryCommand:
    """
    Returns a command for flushing memory writes.
    """
    ...

def getAtomicsProperties(id: int) -> hephaistos.pyhephaistos.AtomicsProperties:
    """
    Returns the atomic capabilities of the device given by its id
    """
    ...

def getCurrentDevice() -> hephaistos.pyhephaistos.Device:
    """
    Returns the currently active device. Note that this may initialize the context.
    """
    ...

def getElementSize(format: hephaistos.pyhephaistos.ImageFormat) -> int:
    """
    Returns the size of a single channel in bytes
    """
    ...

def getEnabledAtomics() -> hephaistos.pyhephaistos.AtomicsProperties:
    """
    Returns the in the current context enabled atomic features
    """
    ...

def getSubgroupProperties(arg: int, /) -> hephaistos.pyhephaistos.SubgroupProperties:
    """
    Returns the properties specific to subgroups (waves).
    """
    ...

@overload
def getSubgroupProperties() -> hephaistos.pyhephaistos.SubgroupProperties:
    """
    Returns the properties specific to subgroups (waves).
    """
    ...

def getSupportedTypes(id: Optional[int], /) -> hephaistos.pyhephaistos.TypeSupport:
    """
    Queries the supported extended types

    Parameters
    ----------
    id: int | None, default=None
        Id of the device to query. If None, uses the one the current context
        was created on
    """
    ...

def isDebugAvailable() -> bool:
    """
    Returns True if debugging is supported on this system.
    This requires the Vulkan Validation Layers to be installed.
    """
    ...

def isDeviceSuitable(arg: int, /) -> bool:
    """
    Returns True if the device given by its id supports all enabled extensions
    """
    ...

def isRaytracingEnabled() -> bool:
    """
    Checks wether ray tracing was enabled. Note that this creates the context.
    """
    ...

def isRaytracingSupported(id: Optional[int] = None) -> bool:
    """
    Checks wether any or the given device supports ray tracing.
    """
    ...

def isVulkanAvailable() -> bool:
    """
    Returns True if Vulkan is available on this system.
    """
    ...

def requireTypes(types: set, force: bool = False, /) -> None:
    """
    Forces the given types as specified by in the set by their names (f64, f16,
    i64, i16, i8), effectively marking devices, which do not support them as
    unsupported. Set force=True if an existing context should be destroyed.
    """

def retrieveImage(
    src: hephaistos.pyhephaistos.Image, dst: hephaistos.pyhephaistos.Buffer
) -> hephaistos.pyhephaistos.RetrieveImageCommand:
    """
    Creates a command for copying the image back into the given buffer.

    Parameters
    ----------
    src: Image
        Source image
    dst: Buffer
        Destination buffer
    """
    ...

def retrieveTensor(
    src: hephaistos.pyhephaistos.Tensor,
    dst: hephaistos.pyhephaistos.Buffer,
    bufferOffset: Optional[int] = None,
    tensorOffset: Optional[int] = None,
    size: Optional[int] = None,
    unsafe: bool = False,
) -> hephaistos.pyhephaistos.RetrieveTensorCommand:
    """
    Creates a command for copying the src tensor back to the destination buffer

    Parameters
    ----------
    src: Tensor
        Source tensor
    dst: Buffer
        Destination buffer
    bufferOffset: None|int, default=None
        Optional offset into the buffer in bytes
    tensorOffset: None|int, default=None
        Optional offset into the tensor in bytes
    size: None|int, default=None
        Amount of data to copy in bytes. If None, equals to the complete buffer
    unsafe: bool, default=False
        Wether to omit barrier ensuring read after write ordering
    """
    ...

def selectDevice(id: int, force: bool = False) -> None:
    """
    Sets the device on which the context will be initialized. Set force=True if
    an existing context should be destroyed.
    """
    ...

def suitableDeviceAvailable() -> bool:
    """
    Returns True, if there is a device available supporting all enabled extensions
    """
    ...

def updateImage(
    src: hephaistos.pyhephaistos.Buffer, dst: hephaistos.pyhephaistos.Image
) -> hephaistos.pyhephaistos.UpdateImageCommand:
    """
    Creates a command for copying data from the given buffer into the image.

    Parameters
    ----------
    src: Buffer
        Source buffer
    dst: Image
        Destination image
    """
    ...

def updateTensor(
    src: hephaistos.pyhephaistos.Buffer,
    dst: hephaistos.pyhephaistos.Tensor,
    bufferOffset: Optional[int] = None,
    tensorOffset: Optional[int] = None,
    size: Optional[int] = None,
    unsafe: bool = False,
) -> hephaistos.pyhephaistos.UpdateTensorCommand:
    """
    Creates a command for copying the src buffer into the destination tensor

    Parameters
    ----------
    src: Buffer
        Source Buffer
    dst: Tensor
        Destination Tensor
    bufferOffset: None|int, default=None
        Optional offset into the buffer in bytes
    tensorOffset: None|int, default=None
        Optional offset into the tensor in bytes
    size: None|int, default=None
        Amount of data to copy in bytes. If None, equals to the complete buffer
    unsafe: bool, default=False
        Wether to omit barrier ensuring read after write ordering
    """
    ...

def updateTexture(
    src: hephaistos.pyhephaistos.Buffer, dst: hephaistos.pyhephaistos.Texture
) -> hephaistos.pyhephaistos.UpdateTextureCommand:
    """
    Creates a command for copying data from the given buffer into the texture.

    Parameters
    ----------
    src: Buffer
        Source buffer
    dst: Texture
        Destination texture
    """
    ...



from __future__ import annotations

from abc import ABC, abstractmethod
from ctypes import Structure, addressof, memmove, sizeof, c_uint8
from itertools import chain
from queue import Queue
from threading import Event, Lock, Thread
import warnings

from hephaistos import (
    Command,
    Program,
    RawBuffer,
    Submission,
    Subroutine,
    Tensor,
    Timeline,
    beginSequence,
    createSubroutine,
    retrieveTensor,
    updateTensor,
)
from hephaistos.util import StructureTensor
from numpy import frombuffer, float32

from numpy.typing import DTypeLike, NDArray
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)


class PipelineStage(ABC):
    """
    Base class for pipeline stages.

    Handles double buffering of configuration, allowing the CPU to update one
    configuration while the other is still in use by the GPU, prevents the GPU
    in the best case from waiting on the CPU, thus boosting throughput.

    UBOs can be defined by passing a dictionary mapping binding names to a
    ctypes structure describing it. It's field name are reflected, added to the
    stages parameters and can be accessed via `getParam` / `setParam`. If they
    start with an underscore, they are considered private and won't show up in
    the `fields` property and output of `getParams` and print.

    Additional properties that should be able to be set via `getParam` /
    `setParam` can be passed by name in the `extras` set. They are most likely
    implemented as properties.

    During creation of a `Pipeline` the method `run(i)` will be called for both
    buffers only once each. It must return a list of commands that define the
    function of the stage and will be used to create a pipeline subroutine that
    gets reused.

    The current idle configuration get's updated via a call to `update(i)`,
    which writes the current local state, i.e. the one defined by the Python
    object, to device memory. Any private parameters can be updated by
    overloading the `_finishParams(i)` method, which gets called during
    `update(i)` before the actual write to the device.

    Parameters
    ----------
    params: {str: Structure}, default={}
        Dictionary of named ctype structure containing the stage's parameters.
        Each structure will be allocated on the CPU side and twice on the GPU
        for buffering. The latter can be bound in programs
    extra: {str}, default={}
        Set of extra parameter name, that can be set and retrieved using the
        stage api. Take precedence over parameters defined by structs. Should
        be be implemented in subclasses as properties.
    """

    name = "stage"
    """default stage name. Should be changed in subclasses."""

    def __init__(
        self, params: Dict[str, Type[Structure]] = {}, extra: Set[str] = set()
    ) -> None:
        # create local configuration
        self._local = {name: param() for name, param in params.items()}
        # create double buffered device config
        self._device = [
            {name: StructureTensor(param, True) for name, param in params.items()}
            for _ in range(2)
        ]
        # check tensors are mapped
        if any(
            not tensor.isMapped for buffer in self._device for tensor in buffer.values()
        ):
            raise RuntimeError("PipelineStage requires support for mapped tensors!")
        # create params map name -> param struct
        self._params = {
            fieldName: fieldType.from_address(
                addressof(self._local[paramName])
                + getattr(params[paramName], fieldName).offset
            )
            for paramName, param in params.items()
            for fieldName, fieldType in param._fields_
            # if not fieldName.startswith("_")
        }
        self._extra = extra
        # collect all fields without private one (starts with "_")
        self._fields = frozenset(extra | self._params.keys())
        self._public = frozenset(f for f in self._fields if not f.startswith("_"))

    @property
    def fields(self) -> Set[str]:
        """Set of all public parameter names"""
        return self._public

    def getParam(self, name: str) -> Any:
        """Returns the parameter specified by its name"""
        if name in self._extra:
            return self.name
        elif name in self._params:
            return self._params[name].value
        else:
            raise ValueError(f"No parameter with name {name}")

    def getParams(self) -> Dict[str, any]:
        """Creates a dictionary with all parameters that can be set"""
        return {name: self.getParam(name) for name in self.fields}

    def setParam(self, name: str, value: Any) -> None:
        """
        If there is a parameter with the given name update its value using the
        provided one, else ignore it.
        """
        if not name in self._fields:
            return
        if name in self._extra:
            setattr(self, name, value)
        elif name in self._params:
            self._params[name].value = value

    def setParams(self, **kwargs) -> None:
        """
        Sets the given parameters in the local configuration.
        Ignores parameters it does not contain.
        """
        for name, value in kwargs.items():
            self.setParam(name, value)

    def __repr__(self) -> str:
        return (
            self.name
            + "\n"
            + "\n".join(f"{param} : {self.getParam(param)}" for param in self.fields)
        )

    def _finishParams(self, i: int) -> None:
        """
        Function called during calls to update().
        Can be overwritten by subclasses to calculate private parameters.
        """
        pass

    def _bindParams(self, program: Program, i: int) -> None:
        """Binds the i-th configuration to the given program"""
        program.bindParams(**self._device[i])

    def update(self, i: int) -> None:
        """
        Updates the i-th configuration stored on the device using the pipeline's
        current state.
        """
        self._finishParams(i)
        for name, struct in self._local.items():
            memmove(self._device[i][name].memory, addressof(struct), sizeof(struct))

    @abstractmethod
    def run(self, i: int) -> List[Command]:
        """
        Creates a list of commands responsible for running the pipeline stage
        using the i-th configuration
        """
        pass


def runPipelineStage(stage: PipelineStage, i: int = 0, *, update: bool = True) -> None:
    """
    Runs the i-th configuration of the given pipeline stage and waits for it to
    finish. If update is True, calls `stage.update(i)` before running it.
    """
    if update:
        stage.update(i)
    beginSequence().AndList(stage.run(i)).Submit().wait()


class SourceCodeMixin(PipelineStage):
    """
    Base class for mixin that inject source code into other stages or programs
    while providing an interface to configure the code in the same way as
    regular pipeline stages.

    Parameters
    ----------
    params: {str: Structure}, default={}
        Dictionary of named ctype structure containing the stage's parameters.
        Each structure will be allocated on the CPU side and twice on the GPU
        for buffering. The latter can be bound in programs
    extra: {str}, default={}
        Set of extra parameter name, that can be set and retrieved using the
        stage api. Take precedence over parameters defined by structs. Should
        be be implemented in subclasses as properties.

    See Also
    --------
    PipelineStage
    """

    def __init__(
        self, params: Dict[str, type[Structure]] = {}, extra: Set[str] = set()
    ) -> None:
        super().__init__(params, extra)

    @property
    @abstractmethod
    def sourceCode(self) -> str:
        """Source code this mixin manages"""
        pass

    def bindParams(self, program: Program, i: int) -> None:
        """
        Binds params used by this mixin to the given program using the i-th
        configuration.
        """
        self._bindParams(program, i)

    def run(self, i: int) -> List[Command]:
        # Most mixin won't add any logic besides the source code
        # so an empty command list is a reasonable default
        return []


class RetrieveTensorStage(PipelineStage):
    """
    Utility stage retrieving a tensor to a local buffer.
    Keeps independent copies for each configuration.
    """

    name = "retrieve"

    def __init__(self, src: Tensor) -> None:
        super().__init__({})
        self._src = src
        # create local buffers
        self._buffers = [RawBuffer(src.size_bytes) for _ in range(2)]
        # create byte array to mimic std::span<std::byte>
        span = c_uint8 * src.size_bytes
        self._data = [span.from_address(buf.address) for buf in self._buffers]

    @property
    def src(self) -> Tensor:
        """Source tensor"""
        return self._src

    def address(self, i: int) -> int:
        """Returns the memory address of the i-th configuration"""
        return self._buffers[i].address

    def buffer(self, i: int) -> RawBuffer:
        """Returns the i-th buffer"""
        return self._buffers[i]

    def view(self, i: int, dtype: DTypeLike = float32) -> NDArray:
        """Interprets the retrieved tensor as array of given type"""
        return frombuffer(self._data[i], dtype)

    def run(self, i: int) -> List[Command]:
        return [retrieveTensor(self._src, self._buffers[i])]


class UpdateTensorStage(PipelineStage):
    """
    Utility stage updating a tensor from a local buffer. Keeps an independent
    source buffer for each configuration to allow concurrent pipeline reading
    and buffer writing.

    To make the buffer update part of the pipeline, consider overwriting
    `_finishParams(config)` in a subclass.
    """

    name = "update"

    def __init__(self, dst: Tensor) -> None:
        super().__init__({})
        self._dst = dst
        # create local buffers
        self._buffers = [RawBuffer(dst.size_bytes) for _ in range(2)]
        # create byte array to mimic std::span<std::byte>
        span = c_uint8 * dst.size_bytes
        self._data = [span.from_address(buf.address) for buf in self._buffers]

    @property
    def dst(self) -> Tensor:
        """Target tensor"""
        return self._dst

    def address(self, i: int) -> int:
        """Returns the memory address of the i-th configuration"""
        return self._buffers[i].address

    def buffer(self, i: int) -> RawBuffer:
        """Returns the i-th buffer"""
        return self._buffers[i]

    def view(self, i: int, dtype: DTypeLike = float32) -> NDArray:
        """Interprets the retrieved tensor as array of given type"""
        return frombuffer(self._data[i], dtype)

    def run(self, i: int) -> List[Command]:
        return [updateTensor(self._buffers[i], self._dst)]


class Pipeline:
    """
    Pipelines contain a sequence of named pipeline stages, manages their states
    as well as providing subroutines to be used for scheduling tasks to be fed
    into the pipeline.
    If no name is provided, the default name of the stage is used. If this name
    is already in use, a increasing number gets added, i.e. the name `stage`
    becomes `stage`, `stage2`, `stage3`, etc.

    Changes to the stages can be issued via a dict assigning a parameter path
    "{stage_name}__{parameter]" to change the parameter in a specific stage or
    just "parameter" to apply the change to all stages with that property.

    Parameters
    ----------
    stages: (PipelineStage | (name, PipelineStage))[]
        sequence of stages the pipeline consists of. Each element can optionally
        specify a name used for updating properties. If no name is provided
        (i.e. not a tuple), it get the name "stage{i}" where i is the stage's
        position in the pipeline
    """

    def __init__(
        self, stages: List[Union[PipelineStage, Tuple[str, PipelineStage]]]
    ) -> None:
        # create stage list and dict
        stages = [s if isinstance(s, tuple) else (s.name, s) for s in stages]
        self._stageList = []  # for ordering
        self._stageDict = {}  # for lookup
        name_counter = {}
        for name, stage in stages:
            # get unique name
            if name in name_counter:
                name_counter[name] += 1
                name = f"{stage.name}{name_counter}"
            else:
                name_counter[name] = 1
            # put stage in list and dict
            self._stageList.append((name, stage))
            self._stageDict[name] = stage

        # create subroutines
        self._subroutines = [
            createSubroutine(
                list(chain.from_iterable(stage.run(i) for _, stage in self._stageList)),
                simultaneous=True,
            )
            for i in range(2)
        ]

    @property
    def stages(self) -> List[Tuple[str, PipelineStage]]:
        """Sequence of named pipeline stages."""
        # create copy to be safe
        return list(self._stageList)

    def getSubroutine(self, i: int) -> Subroutine:
        """
        Returns the subroutine responsible for running the pipeline using
        the i-th configuration
        """
        return self._subroutines[i]

    def getParams(self) -> Dict[str, Any]:
        """
        Collect the parameters of all stages and returns their names and values
        stored in a dict.
        """
        return {
            stageName + "__" + paramName: value
            for stageName, stage in self._stageList
            for paramName, value in stage.getParams().items()
        }

    def setParams(self, **params) -> None:
        """
        Sets the parameters in the stages. A specific stage can be selected via
        `{name}__{parameter}`. If instead only the parameter name is provided,
        the parameter is applied to all stages.
        """
        for name, value in params.items():
            if "__" in name:
                stage, param = name.split("__", 1)
                if not stage in self._stageDict:
                    warnings.warn(f'There is no stage "{stage}" in this pipeline!')
                self._stageDict[stage].setParam(param, value)
            else:
                for _, stage in self._stageList:
                    stage.setParam(name, value)

    def update(self, i: int) -> None:
        """
        Updates the i-th configuration of all stages using their current
        configuration. Note that this does not check if the configuration is
        currently in use and results in undefined behavior if so.
        """
        for _, stage in self._stageList:
            stage.update(i)

    def runAsync(self, i: int, *, update: bool = True) -> Submission:
        """
        Runs the pipeline using the i-th configuration and returns a
        `Submission` which can be used to wait for the pipeline to finish. If
        update is True, updates all stages using their current state before
        running the pipeline. Note that this does not check if the configuration
        is currently in use and results in undefined behavior if so.
        """
        if update:
            self.update(i)
        return beginSequence().And(self.getSubroutine(i)).Submit()

    def run(self, i: int, *, update: bool = True) -> None:
        """
        Runs the pipeline using the i-th configuration and waits for it to
        finish. If update is True, updates all stages using their current state
        before running the pipeline. Note that this does not check if the
        configuration is currently in use and results in undefined behavior if
        so.
        """
        self.runAsync(i, update=update).wait()


def runPipeline(
    stages: Iterable[PipelineStage], i: int = 0, *, update: bool = True
) -> None:
    """
    Runs the i-th configuration of the stages as if they make up a pipeline in
    given order and waits for it to finish. If update is True, calls `update(i)`
    on each stage before running them.

    Calling this function may be faster than creating a `Pipeline`, but is most
    likely slower if called repeatedly. Consider creating a `Pipeline` in that
    case.
    """
    if update:
        for stage in stages:
            stage.update(i)
    beginSequence().AndList(
        list(chain.from_iterable(stage.run(i) for stage in stages))
    ).Submit().wait()


class _CounterWorkerThread:
    """
    Util class for worker threads used by the scheduler. Runs a function as many
    times as specified, while allowing to increase this value. If it is reached
    it suspends the thread ready to run more loops.
    """

    def __init__(self, fn: Callable[[int], None]) -> None:
        self._thread = Thread(target=self._loop, daemon=True)
        self._isSuspending = True
        self._suspendLock = Lock()
        self._wakeEvent = Event()
        self._counter = 0
        self._target = 0
        self._fn = fn
        # start thread
        self._thread.start()

    @property
    def count(self) -> int:
        """Number of finished iterations"""
        return self._counter

    @property
    def total(self) -> int:
        """Total number of issued iterations"""
        return self._target

    def run(self, n: int) -> None:
        """Increases the loop target"""
        self._target += n
        # check if we need to wake
        with self._suspendLock:
            if self._isSuspending:
                self._wakeEvent.set()

    def _loop(self) -> None:
        """Thread body"""
        # infinite loop -> must run inside a daemon thread
        while True:
            # suspend if necessary
            if self._isSuspending:
                self._wakeEvent.wait()
                # resume work
                self._wakeEvent.clear()
                self._isSuspending = False
            # quick and unsafe check
            if self._counter >= self._target:
                # might want to suspend -> check again, but safe
                with self._suspendLock:
                    if self._counter >= self._target:
                        self._isSuspending = True
                        continue
            # run function
            self._fn(self._counter)
            self._counter += 1


class PipelineScheduler:
    """
    Schedules tasks into a pipeline and orchestrates the processing of the
    pipeline results. Bundles tasks into a single batch submission making it
    more efficient than repeatedly calling `run` or `runAsync` on the pipeline.

    Scheduling tasks happens completely in the background using multithreading
    allowing the calling code to do other work, e.g. preparing the next
    submission to be issued to the scheduler, but also provides function to wait
    on the completion of a certain or all tasks.

    New tasks can be issued while previous ones are still processed.

    The scheduler requires exclusive access to two configurations, but may
    otherwise share the pipeline.

    Parameters
    ----------
    pipeline: Pipeline
        Pipeline onto which to schedule tasks
    queueSize: int, default=0
        Size of the task queue. Size of 0 means infinite.
    processFn: Callable( (config: int, task: int) -> None ), default=None
        Function to be called after each finished task on the pipeline with
        config being the pipeline's config used and task the increasing task
        count processed, i.e. 0 for the first task, 1 for the second and so on.
        The scheduler ensures tasks using the same configuration to wait until
        processing finished. Will run in its own thread.
    """

    def __init__(
        self,
        pipeline: Pipeline,
        *,
        queueSize: int = 0,
        processFn: Optional[Callable[[int, int], None]] = None,
    ) -> None:
        # save params
        self._queueSize = queueSize
        self._totalTasks = 0
        self._pipeline = pipeline
        self._pipelineTimeline = Timeline()
        # create queue worker
        self._updateTimeline = Timeline()
        self._updateWorker = _CounterWorkerThread(self._update)
        self._updateQueue = Queue()
        # create process worker if needed
        self._processFn = processFn
        if processFn is None:
            self._processWorker = None
            self._processTimeline = None
        else:
            self._processWorker = _CounterWorkerThread(self._process)
            self._processTimeline = Timeline()

    @property
    def pipeline(self) -> Pipeline:
        """The underlying pipeline into which tasks are orchestrated"""
        return self._pipeline

    @property
    def queueSize(self) -> int:
        """
        Capacity of the internal queue used for scheduling tasks.
        A value of zero means infinite.
        """
        return self._queueSize

    @property
    def totalTasks(self) -> int:
        """Total number of tasks scheduled"""
        return self._totalTasks

    @property
    def tasksScheduled(self) -> int:
        """Approximate number of tasks scheduled"""
        return self._updateQueue.qsize()

    @property
    def tasksFinished(self) -> int:
        """Returns the number of finished tasks processed by the timeline"""
        if self._processWorker is None:
            return self._pipelineTimeline.value
        else:
            return self._processWorker.count

    def schedule(
        self, tasks: Iterable[Dict[str, Any]], *, timeout: Optional[float] = None
    ) -> Tuple[int, Submission]:
        """
        Schedules the given list of tasks to be processed by the pipeline after
        previous tasks submissions have finished.

        Parameters
        ----------
        tasks: Iterable[Dict[str, Any]]
            Sequence of tasks to schedule onto the pipeline defined by the set
            of parameters to apply to the stages.
        timeout: float | None, default=None
            Timeout in seconds for waiting on free space in the queue.
            If None, waits indefinitely.

        Returns
        -------
        nSubmitted: int
            Number of tasks actually submitted.
        submission: Submission | None
            The submission created for submitting work to the GPU. Can be used
            to wait on the tasks to finish or query the final task index via
            `submission.finalStep`.
            None if nSubmitted == 0.
        """
        # put task onto queue
        n = 0
        builder = None
        for task in tasks:
            # try to enlist in queue
            try:
                self._updateQueue.put(task, timeout=timeout)
            except:
                break

            # lazy create builder
            if builder is None:
                builder = beginSequence(self._pipelineTimeline, self._totalTasks)

            # issue wait on previous task
            builder.WaitFor(self._pipelineTimeline, self._totalTasks)
            # issue wait on update thread
            builder.WaitFor(self._updateTimeline, self._totalTasks + 1)
            # issue wait on process thread is present
            if self._processTimeline is not None and self._totalTasks >= 2:
                # double buffered -> wait for the processing of tasks scheduled
                # two earlier to finish to make sure we don't overwrite the
                # result it tries to process
                builder.WaitFor(self._processTimeline, self._totalTasks - 1)
            # run task
            i = self._totalTasks % 2
            builder.And(self._pipeline.getSubroutine(i))

            # update counters
            n += 1
            self._totalTasks += 1

        # enqueued anything?
        if n == 0:
            return (0, None)

        # submit work
        submission = builder.Submit()
        # just to ensure we're not breaking stuff in the future
        assert submission.forgettable

        # update worker threads
        self._updateWorker.run(n)
        if self._processWorker is not None:
            self._processWorker.run(n)

        # return result
        return (n, submission)

    def wait(self, task: Optional[int] = None) -> None:
        """
        Waits on the given task to finish. Blocks the calling code.
        If task is None, waits on the last scheduled task.

        Waiting on a task not scheduled may result in a deadlock if no further
        thread schedule tasks.
        """
        if task is None:
            task = self.totalTasks
        if self._processTimeline is None:
            self._pipelineTimeline.wait(task)
        else:
            self._processTimeline.wait(task)

    def waitTimeout(self, task: Optional[int] = None, *, timeout: int = 1000) -> bool:
        """
        Waits on the given task to finish for at most `timeout` nanoseconds.

        Parameters
        ----------
        task: int | None, default=None
            Position of the task to wait on. Waits on the last added one if None.
        timeout: int, default=1000
            Timeout in nanoseconds. The driver is free to round this value up to
            its internal resolution. May be zero to query the tasks state
            immediately.

        Returns
        -------
        finished: bool
            True, if the given task finished.
        """
        if task is None:
            task = self.totalTasks
        if self._processTimeline is None:
            self._pipelineTimeline.waitTimeout(task, timeout)
        else:
            self._processTimeline.waitTimeout(task, timeout)

    def _update(self, n: int) -> None:
        """Internal update thread body"""
        # fetch next task (don't need to block, as this is handled by the worker)
        task: Optional[Dict[str, Any]] = self._updateQueue.get(block=False)
        # update pipeline
        self._pipeline.setParams(**task)
        # wait for the i-th config to be safe to update
        if n >= 2:
            self._pipelineTimeline.wait(n - 1)
        # update config
        # eventually calls user provided functions
        # -> treat as evil to prevent from deadlocking timeline
        try:
            self._pipeline.update(n % 2)
        except Exception as ex:
            warnings.warn(f"Exception raised while preparing task {n}:\n{ex}")
        # advance timeline
        self._updateTimeline.value = n + 1

    def _process(self, n: int) -> None:
        """Internal process thread body"""
        # wait on task to finish
        self._pipelineTimeline.wait(n + 1)
        # process
        # external provided function
        # -> treat as evil to prevent from deadlocking timeline
        try:
            self._processFn(n % 2, n)
        except Exception as ex:
            warnings.warn(f"Exception raised while processing task {n}:\n{ex}")
        # advance timeline
        self._processTimeline.value = n + 1



import ctypes
from typing import Any, Union
from numpy.ctypeslib import as_array
from .pyhephaistos import ByteTensor, RawBuffer


def printSize(size_bytes: int) -> str:
    """
    Creates a printable representation of the given size in bytes by using a
    sensible unit like KB or MB.
    """
    if size_bytes == 0:
        return "0B"
    names = ["B", "KB", "MB", "GB", "TB", "PB"]
    dim = (size_bytes.bit_length() - 1) // 10
    dim = min(dim, len(names) - 1)
    unit = 2 << (dim * 10) - 1
    return f"{size_bytes/unit:.1f} {names[dim]}"


def collectFields(struct: ctypes.Structure) -> list[tuple[str, ctypes._SimpleCData]]:
    """
    Returns a flat list of fields and their types for a given ctype structure.
    """
    fields = []
    for name, T in struct._fields_:
        if type(T) == type(ctypes.Structure):
            fields.extend(
                [(f"{name}.{subname}", subT) for subname, subT in collectFields(T)]
            )
        elif type(T) == type(ctypes.Array):
            # array of structs or simple types?
            if type(T._type_) == type(ctypes.Structure):
                fields.extend(
                    [
                        (f"{name}[{i}].{subname}", subT)
                        for i in range(T._length_)
                        for subname, subT in collectFields(T._type_)
                    ]
                )
            else:
                fields.extend([(f"{name}[{i}]", T._type_) for i in range(T._length_)])
        else:
            fields.append((name, T))
    return fields


def createFlatType(struct: ctypes.Structure) -> ctypes.Structure:
    """
    Creates a equivalent ctype structure from the given one where all nested
    types are flattened.
    """
    fields = collectFields(struct)

    class Flat(ctypes.Structure):
        _fields_ = fields

    return Flat


class ArrayBuffer(RawBuffer):
    """
    Helper class for creating buffer holding an array of a given type.
    """

    def __init__(self, type: ctypes.Structure, size: int):
        """
        Creates a typed array buffer of the given type and size.
        """
        super().__init__(ctypes.sizeof(type) * size)
        self._arr = (type * size).from_address(super().address)
        self._flat = (createFlatType(type) * size).from_address(super().address)

    @property
    def array(self):
        return self._arr

    @property
    def flatarray(self):
        return self._flat

    def numpy(self, flat: bool = False):
        """
        Returns a structured numpy array sharing the same memory as the buffer.
        If flat is True, the field names correspond the the flattened ones, i.e.
        ["dir.x"] instead of ["dir"]["x"].
        """
        return as_array(self.flatarray if flat else self.array)

    def __len__(self):
        return self._arr.__len__()

    def __getitem__(self, key):
        return self._arr.__getitem__(key)

    def __setitem__(self, key, item):
        return self._arr.__setitem__(key, item)

    def __iter__(self):
        return iter(self._arr)


class ArrayTensor(ByteTensor):
    """
    Helper class for creating a tensor big enough to hold an array of given type
    and size.
    """

    def __init__(self, type: ctypes.Structure, size: int, mapped: bool = False):
        """
        Create a typed array tensor of given type and size.
        """
        super().__init__(ctypes.sizeof(type) * size, mapped)


class StructureBuffer(RawBuffer):
    """
    Helper class for creating typed buffers.
    """

    def __init__(self, type: ctypes.Structure):
        """
        Creates a typed buffer with the given type.
        """
        super().__init__(ctypes.sizeof(type))
        # bypass __setattr__ we are overriding
        super().__setattr__(
            "_pointer", ctypes.cast(super().address, ctypes.POINTER(type))
        )

    def __getattribute__(self, __name: str):
        # special types: size, size_bytes
        if __name in ["size", "size_bytes"]:
            return super().__getattribute__(__name)
        # bypass __getattribute__ to prevent infinite loop
        return super().__getattribute__("_pointer").contents.__getattribute__(__name)

    def __setattr__(self, __name: str, __value: Any) -> None:
        # bypass our own __getattribute__
        super().__getattribute__("_pointer").contents.__setattr__(__name, __value)


class StructureTensor(ByteTensor):
    """
    Helper class for creating tensors matching the size of a given struct.
    """

    def __init__(self, typeOrData: Union[ctypes.Structure, Any], mapped: bool = False):
        """
        Creates a tensor with the exact size to hold the given structure.
        """
        if type(typeOrData) == type(ctypes.Structure):
            super().__init__(ctypes.sizeof(typeOrData), mapped)
        else:
            super().__init__(
                ctypes.addressof(typeOrData), ctypes.sizeof(typeOrData), mapped
            )



from __future__ import annotations
import numpy as np
from ctypes import Structure, c_float, c_uint32, c_int32
from typing import Type, Union
from hephaistos import Tensor


class vec(Structure):
    """Base class for GLSL vectors. Supports array indices and swizzling"""

    # def __new__(cls, *arg, **kwargs) -> Self:
    #     if hasattr(cls, "_fields_"):
    #         return super().__new__(cls, kwargs)
    #     if not 2 <= cls._dim_ <= 4:
    #         raise ValueError("Vector dimension must be 2,3 or 4!")
    #     cls._fields_ = [(d, cls._scalar_) for d in ["x", "y", "z", "w"][: cls._dim_]]
    #     return super().__new__(cls, **kwargs)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if len(args) == 1:
            self[:] = args[0]
        elif len(args) > 1:
            self[:] = args

    def __len__(self) -> int:
        return self._dim_

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return tuple(self[i] for i in range(self._dim_)[idx])
        elif idx >= self._dim_:
            raise IndexError("vector has no such dimension")
        else:
            return getattr(self, ["x", "y", "z", "w"][idx])

    def __getattr__(self, name: str):
        if len(name) == 1:
            return super().__getattr__(name)
        else:
            return tuple(getattr(self, atr) for atr in name)

    def __setitem__(self, idx, value):
        if isinstance(idx, slice):
            for i in range(self._dim_)[idx]:
                try:
                    self[i] = value[i]
                except:
                    self[i] = value
        elif idx >= self._dim_:
            raise IndexError("vector has no such dimension")
        else:
            return setattr(self, ["x", "y", "z", "w"][idx], value)

    def __setattr__(self, name: str, value) -> None:
        # special case: value property
        if name == "value":
            self[:] = value
        if len(name) == 1:
            return super().__setattr__(name, value)
        for i, atr in enumerate(name):
            try:
                super().__setattr__(atr, value[i])
            except:
                super().__setattr__(atr, value)

    @property
    def value(self):
        return tuple(self[i] for i in range(self._dim_))

    @value.setter
    def value(self, value):
        self[:] = value

    def __repr__(self) -> str:
        return str(tuple(self))


# Had some weird bug, where vec were constructed without any fields assigned
# this is a work around, but still have no idea why it's needed
def _initVecFields(scalar, dim):
    return [(["x","y","z","w"][d], scalar) for d in range(dim)]

class vec2(vec):
    _fields_ = _initVecFields(c_float, 2)
    _scalar_ = c_float
    _dim_ = 2


class vec3(vec):
    _fields_ = _initVecFields(c_float, 3)
    _scalar_ = c_float
    _dim_ = 3


class vec4(vec):
    _fields_ = _initVecFields(c_float, 4)
    _scalar_ = c_float
    _dim_ = 4


class ivec2(vec):
    _fields_ = _initVecFields(c_int32, 2)
    _scalar_ = c_int32
    _dim_ = 2


class ivec3(vec):
    _fields_ = _initVecFields(c_int32, 3)
    _scalar_ = c_int32
    _dim_ = 3


class ivec4(vec):
    _fields_ = _initVecFields(c_int32, 4)
    _scalar_ = c_int32
    _dim_ = 4


class uvec2(vec):
    _fields_ = _initVecFields(c_uint32, 2)
    _scalar_ = c_uint32
    _dim_ = 2


class uvec3(vec):
    _fields_ = _initVecFields(c_uint32, 3)
    _scalar_ = c_uint32
    _dim_ = 3


class uvec4(vec):
    _fields_ = _initVecFields(c_uint32, 4)
    _scalar_ = c_uint32
    _dim_ = 4


class mat(Structure):
    """Base class for GLSL matrices"""

    # def __new__(cls, *args) -> Self:
    #     if not hasattr(cls, "_fields_"):
    #         cls._fields_ = [("items", cls._type_ * cls._length_)]
    #     return super().__new__(cls)

    def __init__(self, *args):
        super().__init__()
        if len(args) == 1:
            self[:] = args[0]
        elif len(args) > 1:
            self[:] = args

    def __len__(self) -> int:
        return self._length_

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return tuple(self[i] for i in range(self._length_)[idx])
        return self.items[idx]

    def __setitem__(self, idx, value):
        if isinstance(idx, slice):
            for i in range(self._length_)[idx]:
                try:
                    self[i][:] = value[i]
                except:
                    self[i][:] = value
        else:
            self.items[idx][:] = value

    @property
    def value(self):
        return tuple(self[i] for i in range(self._length_))

    @value.setter
    def value(self, value):
        self[:] = value

    def __repr__(self) -> str:
        return "\n".join(str(self.items[i]) for i in range(self._length_))


class mat2(mat):
    _fields_ = [("items", vec2*2)]
    _type_ = vec2
    _length_ = 2


class mat3x2(mat):
    _fields_ = [("items", vec2*3)]
    _type_ = vec2
    _length_ = 3


class mat4x2(mat):
    _fields_ = [("items", vec2*4)]
    _type_ = vec2
    _length_ = 4


class mat2x3(mat):
    _fields_ = [("items", vec3*2)]
    _type_ = vec3
    _length_ = 3


class mat3(mat):
    _fields_ = [("items", vec3*3)]
    _type_ = vec3
    _length_ = 3


class mat4x3(mat):
    _fields_ = [("items", vec3*4)]
    _type_ = vec3
    _length_ = 4


class mat2x4(mat):
    _fields_ = [("items", vec4*2)]
    _type_ = vec4
    _length_ = 2


class mat3x4(mat):
    _fields_ = [("items", vec4*3)]
    _type_ = vec4
    _length_ = 3


class mat4(mat):
    _fields_ = [("items", vec4*4)]
    _type_ = vec4
    _length_ = 4


def stackVector(arrays, vecType: Type[vec]):
    """
    Stacks arrays to produce an array of type vecType, i.e. each array provides
    the data for the dimension corresponding to its index.

    Parameters
    ----------
    array: sequence of array_like
        Data for each dimension of the vectors

    vecType: Vector type
        Type of each column in the resulting array

    Returns
    -------
    stacked: ndarray, dtype=vecType
        The stacked array of given vector type

    Examples
    --------
    >>> x = np.array([1.0,2.0])
    >>> y = np.array([3.0,4.0])
    >>> z = np.array([5.0,6.0])
    >>> data = stackVector((x,y,z), vec3)
    >>> data
    array([(1., 3., 5.), (2., 4., 6.)],
      dtype={'names': ['x', 'y', 'z'],
      'formats': ['<f4', '<f4', '<f4'],
      'offsets': [0, 4, 8],
      'itemsize': 12, 'aligned': True})

    >>> buf = ArrayBuffer(vec3, 2)
    >>> buf.numpy()[:] = data
    >>> buf
    array([(1., 3., 5.), (2., 4., 6.)],
      dtype=[('x', '<f4'), ('y', '<f4'), ('z', '<f4')])
    """
    return (
        np.stack(arrays, axis=-1, dtype=np.dtype(vecType._scalar_))
        .ravel()
        .view(dtype=vecType)
    )


class buffer_reference(Structure):
    """packed buffer reference into a uvec2"""

    _fields_ = [("lo", c_uint32), ("hi", c_uint32)]

    def __init__(self, value=0):
        self.value = value

    @property
    def value(self) -> int:
        return self.lo + (self.hi << 32)

    @value.setter
    def value(self, value: Union[int, Tensor]) -> None:
        if isinstance(value, Tensor):
            value = value.address
        self.lo = value & 0xFFFFFFFF
        self.hi = value >> 32 & 0xFFFFFFFF



from .pyhephaistos import *
from .util import (
    ArrayBuffer,
    ArrayTensor,
    StructureBuffer,
    StructureTensor,
)

# make the Program.bindParams more pythonic
def _bindParams(program: Program, *params, **namedparams) -> None:
    """
    Helper function to bind a list of params in a program.
    """
    for i, p in enumerate(params):
        p.bindParameter(program, i)
    
    names = [b.name for b in program.bindings]
    for name, p in namedparams.items():
        if name not in names:
            continue
        p.bindParameter(program, name)
#register function in class
Program.bindParams = _bindParams
